/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "ICMPCrafter.h"

#include "ErrorHandlerMacros.h"
#include "Array_ptr.h"

#include <memory>

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

KOMODIA_NAMESPACE_START

#define CICMPCrafter_Class "CICMPCrafter"

CICMPCrafter::CICMPCrafter() : CIPCrafter(),
							   CICMPSocket(TRUE)
{
	try
	{
		//Set our name
		SetName(CICMPCrafter_Class);

		//Set the protocol
		SetProtocol(IPPROTO_ICMP);

		//Data initialization
		m_bDefaultChecksum=FALSE;
		m_usChecksum=0;
	}
	ERROR_HANDLER("CICMPCrafter")
}	

CICMPCrafter::~CICMPCrafter()
{
}

LPIpHeader CICMPCrafter::ConstructIPHeader(unsigned char  ucProtocol,
										   unsigned short usFragmentationFlags,
										   unsigned char  ucTTL,
										   unsigned short usIdentification,
										   unsigned char  ucHeaderLength)const
{
	return ConstructCraftedIPHeader(ucProtocol,
									usFragmentationFlags,
									ucTTL,
									usIdentification,
									ucHeaderLength);
}

void CICMPCrafter::FinalIPHeader(LPIpHeader lpHead)const
{
	try
	{
		//Check if we need to modify the header
		if (!m_DefaultChecksum)
			lpHead->usChecksum=htons(m_usChecksum);
	}
	ERROR_HANDLER("FinalIPHeader")
}

void CICMPCrafter::SetICMPUseDefaultChecksum(BOOL bDefault)
{
	m_bDefaultChecksum=bDefault;
}

void CICMPCrafter::SetICMPChecksum(unsigned short usChecksum)
{
	m_usChecksum=usChecksum;
}

void CICMPCrafter::FinalICMPHeader(LPICMPHeader lpHead)const
{
	if (!m_bDefaultChecksum)
		lpHead->usICMPChecksum=htons(m_usChecksum);
}

void CICMPCrafter::SetDestinationAddress(IP aDestinationAddress)
{
	m_aDestinationAddress=aDestinationAddress;
}

void CICMPCrafter::SetDestinationAddress(const std::string& rDestinationAddress)
{
	//Check if null
	if (rDestinationAddress.empty())
		m_aDestinationAddress=0;
	else
		m_aDestinationAddress=StringToLong(rDestinationAddress);
}

char* CICMPCrafter::GetTime(unsigned char cType,
						    unsigned short& usTotalSize)const
{
	try
	{
		return CreateFinal(ICMP_Time,
						   cType,
						   usTotalSize);
	}
	ERROR_HANDLER_RETURN("GetTime",NULL)
}

char* CICMPCrafter::GetParameter(unsigned char cError,
								 unsigned short& usTotalSize)const
{
	try
	{
		//Consturct the packet
		LPICMPHeader lpHead;
		lpHead=ConstructICMP();

		if (!lpHead)
		{
			ReportError("GetParameter","Failed to construct ICMP header!");
			return FALSE;
		}

		//Protect it
		std::auto_ptr<ICMPHeader> pProtection(lpHead);

		//Set the values
		lpHead->ucICMPType=ICMP_Parameter;
		lpHead->ucICMPCode=ICMP_Parameter_ERROR;
		lpHead->sICMP.sUC.uc1=cError;

		//And the checksum
		lpHead->usICMPChecksum=CalculateChecksum((unsigned short*)lpHead,ICMPHeaderLength);

		//Create the packet
		return CreateFinal(lpHead,
						   usTotalSize);
	}
	ERROR_HANDLER_RETURN("GetParameter",NULL)
}

char* CICMPCrafter::GetQuench(unsigned short& usTotalSize)const
{
	try
	{
		return CreateFinal(ICMP_Quench,
						   0,
						   usTotalSize);
	}
	ERROR_HANDLER_RETURN("GetQuench",NULL)
}

char* CICMPCrafter::GetRedirect(unsigned char cType, 
							    IP aGatewayAddress,
								unsigned short& usTotalSize)const
{
	try
	{
		//Construct the header
		LPICMPHeader lpHead;
		lpHead=ConstructICMP();

		if (!lpHead)
		{
			ReportError("GetRedirect","Failed to construct ICMP header!");
			return FALSE;
		}

		//Protect it
		std::auto_ptr<ICMPHeader> pProtection(lpHead);

		//Set the values
		lpHead->ucICMPType=ICMP_Redirect;
		lpHead->ucICMPCode=cType;
		lpHead->sICMP.sUL=aGatewayAddress;

		//And the checksum
		lpHead->usICMPChecksum=CalculateChecksum((unsigned short*)lpHead,ICMPHeaderLength);

		//Create the packet
		return CreateFinal(lpHead,
						   usTotalSize);
	}
	ERROR_HANDLER_RETURN("GetRedirect",NULL)
}

char* CICMPCrafter::GetEcho(BOOL bReply, 
						    unsigned short usIdentifier, 
						    unsigned short usSequence, 
						    unsigned long ulData,
							unsigned short& usTotalSize)const
{
	try
	{
		//Create the header
		LPICMPHeader lpHead;
		lpHead=ConstructICMP();
		
		if (!lpHead)
		{
			ReportError("SendEcho","Failed to construct ICMP header!");
			return FALSE;
		}

		//Protect it
		std::auto_ptr<ICMPHeader> pProtection(lpHead);

		//Check if echo or reply
		if (bReply)
			lpHead->ucICMPType=ICMP_Echo_Reply;
		else
			lpHead->ucICMPType=ICMP_Echo;

		lpHead->ucICMPCode=0;
		lpHead->sICMP.sUS.us1=htons(usIdentifier);
		lpHead->sICMP.sUS.us2=htons(usSequence);
		lpHead->ulICMP_Originate_Timestamp=htonl(ulData);

		//And the checksum
		lpHead->usICMPChecksum=CalculateChecksum((unsigned short*)lpHead,ICMPHeaderLength);

		//Create the packet
		return CreateFinal(lpHead,
						   usTotalSize);
	}
	ERROR_HANDLER_RETURN("GetEcho",NULL)
}

char* CICMPCrafter::GetTimestamp(BOOL bReply, 
								 unsigned short usIdentifier, 
								 unsigned short usSequence, 
								 unsigned long ulOriginateTimestamp, 
								 unsigned long ulReceiveTimestamp, 
								 unsigned long ulTransmitTimestamp,
								 unsigned short& usTotalSize)const
{
	try
	{
		//Create the header
		LPICMPHeader lpHead;
		lpHead=ConstructICMP();

		if (!lpHead)
		{
			ReportError("GetTimestamp","Failed to construct ICMP header!");
			return FALSE;
		}

		//Protect it
		std::auto_ptr<ICMPHeader> pProtection(lpHead);

		//Check if echo or reply
		if (bReply)
			lpHead->ucICMPType=ICMP_Timestamp_Reply;
		else
			lpHead->ucICMPType=ICMP_Timestamp;

		lpHead->ucICMPCode=0;
		lpHead->sICMP.sUS.us1=htons(usIdentifier);
		lpHead->sICMP.sUS.us2=htons(usSequence);
		lpHead->ulICMP_Originate_Timestamp=htonl(ulOriginateTimestamp);
		lpHead->ulICMP_Receive_Timestamp=htonl(ulReceiveTimestamp);
		lpHead->ulICMP_Transmit_Timestamp=htonl(ulTransmitTimestamp);

		//And the checksum
		lpHead->usICMPChecksum=CalculateChecksum((unsigned short*)lpHead,ICMPHeaderLength);

		//Create the packet
		return CreateFinal(lpHead,
						   usTotalSize);
	}
	ERROR_HANDLER_RETURN("GetTimestamp",NULL)
}

char* CICMPCrafter::GetInformation(BOOL bReply, 
								   unsigned short usIdentifier, 
								   unsigned short usSequence,
								   unsigned short& usTotalSize)const
{
	try
	{
		//Create the header
		LPICMPHeader lpHead;
		lpHead=ConstructICMP();

		if (!lpHead)
		{
			ReportError("GetInformation","Failed to construct ICMP header!");
			return FALSE;
		}

		//Protect it
		std::auto_ptr<ICMPHeader> pProtection(lpHead);

		//Check if echo or reply
		if (bReply)
			lpHead->ucICMPType=ICMP_Information_Reply;
		else
			lpHead->ucICMPType=ICMP_Information;

		lpHead->ucICMPCode=0;
		lpHead->sICMP.sUS.us1=htons(usIdentifier);
		lpHead->sICMP.sUS.us2=htons(usSequence);
		
		//And the checksum
		//Using only first 8 bytes
		lpHead->usICMPChecksum=CalculateChecksum((unsigned short*)lpHead,ICMP_Information_SIZE);

		//Create the packet
		return CreateFinal(lpHead,
						   usTotalSize);
	}
	ERROR_HANDLER_RETURN("GetInformation",NULL)
}

char* CICMPCrafter::GetUnreachable(unsigned char cType,
								   unsigned short& usTotalSize)const
{
	try
	{
		return CreateFinal(ICMP_Unreachable,
						   cType,
						   usTotalSize);
	}
	ERROR_HANDLER_RETURN("GetUnreachable",FALSE)
}

char* CICMPCrafter::CreateFinal(unsigned char cICMP, 
							    unsigned char cType,
							    unsigned short& usTotalSize)const
{
	try
	{
		//Generic ICMP send
		LPICMPHeader lpHead;
		lpHead=ConstructICMP();

		if (!lpHead)
		{
			ReportError("CreateFinal","Failed to construct ICMP header!");
			return NULL;
		}

		//Protect it
		std::auto_ptr<ICMPHeader> pProtection(lpHead);

		//Set the data
		lpHead->ucICMPType=cICMP;
		lpHead->ucICMPCode=cType;

		//And the checksum
		lpHead->usICMPChecksum=CalculateChecksum((unsigned short*)lpHead,ICMPHeaderLength);

		//Call the final modifier
		FinalICMPHeader(lpHead);

		//And construct higher level
		return CIPCrafter::GetCraftedPacket(this,
											m_aDestinationAddress,
											(char*)lpHead,
											ICMPHeaderLength,
											usTotalSize);
	}
	ERROR_HANDLER_RETURN("CreateFinal",NULL)
}

char* CICMPCrafter::CreateFinal(const LPICMPHeader pHeader,
								unsigned short& usTotalSize)const
{
	try
	{
		//Do we have a send buffer
		if (GetSendBuffer())
		{
			//Calculate the new size
			unsigned long ulNewSize;
			ulNewSize=GetBufferSize()+ICMPHeaderLength;

			//Create a new buffer for it
			char* pBuffer;
			pBuffer=new char[ulNewSize];

			//Protect it
			CArray_ptr<char> pProtection(pBuffer);

			//Copy the data
			memcpy(pBuffer,pHeader,ICMPHeaderLength);
			memcpy(pBuffer+ICMPHeaderLength,
				   GetSendBuffer(),
				   GetBufferSize());

			//And create the crafted packet
			return CIPCrafter::GetCraftedPacket(this,
												m_aDestinationAddress,
												pBuffer,
												ulNewSize,
												usTotalSize);
		}
		else
			return CIPCrafter::GetCraftedPacket(this,
												m_aDestinationAddress,
												(char*)pHeader,
												ICMPHeaderLength,
												usTotalSize);
	}
	ERROR_HANDLER_RETURN("CreateFinal",NULL)
}

KOMODIA_NAMESPACE_END
