/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "DNSAnswers.h"

#include "ErrorHandlerMacros.h"
#include "Array_ptr.h"

#include "SpoofBase.h"

#include "DNSParser.h"

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

namespace KomodiaDNS
{

#define CDNSAnswers_Class "CDNSAnswers"

CDNSAnswers::CDNSAnswers() : CErrorHandler(),
							 m_pDNSQuery(NULL)
{
	try
	{
		//Set our name
		SetName(CDNSAnswers_Class);
	}
	ERROR_HANDLER("CDNSAnswers")
}

CDNSAnswers::CDNSAnswers(const CDNSAnswers& rAnswers) : CErrorHandler(),
														m_pDNSQuery(NULL),
														m_aAnswers(rAnswers.m_aAnswers)
{
	try
	{
		//Set our name
		SetName(CDNSAnswers_Class);

		//Do we have a query?
		if (rAnswers.m_pDNSQuery)
			//Copy it
			m_pDNSQuery=new CDNSQuery(*rAnswers.m_pDNSQuery);
	}
	ERROR_HANDLER("CDNSAnswers")
}

CDNSAnswers::~CDNSAnswers()
{
	try
	{
		//Delete the query
		delete m_pDNSQuery;
	}
	ERROR_HANDLER("~CDNSAnswers")
}

const CDNSAnswer& CDNSAnswers::GetAnswer(unsigned short usIndex)const
{
	return m_aAnswers[usIndex];
}

int CDNSAnswers::Parse(const char *pBuffer,
					   int iLength)
{
	try
	{
		//Delete previous data
		UnInitialize();

		if (iLength<=DnsHeaderHeaderLength)
		{
			//Report it
			ReportError("Parse","Buffer size too small!");

			//Exit
			return 0;
		}

		//Start parsing the dude
		const char* pBackupBuffer=pBuffer;

		//Let's start with the answer
		memcpy(&m_aDNSHeader,
			   pBackupBuffer,
			   DnsHeaderHeaderLength);

		//Reverse the header
		ReverseHeader();

		//Increase the buffer
		pBackupBuffer+=DnsHeaderHeaderLength;

		//OK now we can continue
		m_pDNSQuery=new CDNSQuery;

		//Get the number of data the query parsed !
		unsigned short usParseLength;

		//Parse it
		usParseLength=m_pDNSQuery->Parse(pBackupBuffer,
									     m_aDNSHeader.usQDCount);

		//Check for errors
		if (!usParseLength)
		{
			//Report it
			ReportError("Parse","Received zero length!");

			//Exit
			return 0;
		}

		//Increase the buffer
		pBackupBuffer+=usParseLength;

		//Parse the rest of the headers
		return ParseAnswers(pBackupBuffer,
							pBuffer);
	}
	ERROR_HANDLER_RETURN("Parse",0)
}

void CDNSAnswers::UnInitialize()
{
	try
	{
		//Delete the query
		delete m_pDNSQuery;
		m_pDNSQuery=NULL;

		//Clear the vector
		m_aAnswers.clear();
	}
	ERROR_HANDLER("UnInitialize")
}

void CDNSAnswers::ReverseHeader()
{
	try
	{
		//Everything is reversed, change it
		m_aDNSHeader.usANCount=htons(m_aDNSHeader.usANCount);
		m_aDNSHeader.usARCount=htons(m_aDNSHeader.usARCount);
		m_aDNSHeader.usID=htons(m_aDNSHeader.usID);
		m_aDNSHeader.usNSCount=htons(m_aDNSHeader.usNSCount);
		m_aDNSHeader.usQDCount=htons(m_aDNSHeader.usQDCount);
	}
	ERROR_HANDLER("ReverseHeader")
}

unsigned short CDNSAnswers::ParseAnswers(const char *pBuffer,
										 const char* pOriginalBuffer)
{
	try
	{
		//Let's start
		//Get number of answers
		int iAnswers;
		iAnswers=m_aDNSHeader.usARCount+
				 m_aDNSHeader.usANCount+
				 m_aDNSHeader.usNSCount;

		for (int iCount=0;iCount<iAnswers;++iCount)
		{
			//Our answer
			CDNSAnswer aAnswer;

			//Try to parse it
			unsigned short usParseSize;
			usParseSize=aAnswer.ParseBuffer(pBuffer,
											pOriginalBuffer);

			if (!usParseSize)
			{
				//Report it
				ReportError("ParseAnswers","Failed to parse answer!");

				//Exit
				return 0;
			}

			//Set the answer type
			CDNSAnswer::DnsHeaderType aAnswerType;
			if (iCount<m_aDNSHeader.usANCount)
				aAnswerType=CDNSAnswer::dhAnswer;
			else if (iCount<m_aDNSHeader.usANCount+
							m_aDNSHeader.usNSCount)
				aAnswerType=CDNSAnswer::dhAuthoritive;
			else
				aAnswerType=CDNSAnswer::dhAdditional;

			//Set it at the class
			aAnswer.SetAnswerType(aAnswerType);

			//Add it to the vector
			m_aAnswers.push_back(aAnswer);

			//Increase the buffer size
			pBuffer+=usParseSize;
		}

		//Done
		return pBuffer-pOriginalBuffer;
	}
	ERROR_HANDLER_RETURN("ParseAnswers",0)
}

unsigned long CDNSAnswers::GetAddress(unsigned short usIndex)const
{
	try
	{
		if (!m_aDNSHeader.usARCount || 
			!m_aDNSHeader.usANCount)
		{
			//Report it
			ReportError("GetAddress","No answers!");

			//Exit
			return 0;
		}

		//Get to the actual answer
		unsigned short usPrior;
		usPrior=m_aDNSHeader.usANCount+
			    m_aDNSHeader.usNSCount;

		for (unsigned short usCounter=usIndex;
			 usCounter<m_aDNSHeader.usARCount-usIndex;
			 ++usCounter)
			if (m_aAnswers[usPrior+usCounter].GetResourceType()==CDNSQuery::A)
			{
				//Get the data
				std::string sData;
				sData=m_aAnswers[usPrior+usCounter].GetData();

				//Convert it to long
				return atol(sData.c_str());
			}

		//Nothing
		return 0;
	}
	ERROR_HANDLER_RETURN("GetAddress",0)
}

BOOL CDNSAnswers::IsResolved()const
{
	return !(m_aDNSHeader.usOptions & DNS_SERVER_FAILURE);
}

unsigned short CDNSAnswers::GetDNSID() const
{
	return m_aDNSHeader.usID;
}

unsigned long CDNSAnswers::GetAIPAnswer()const
{
	try
	{
		//Do we have an answer ?
		if (!m_aDNSHeader.usANCount ||
			m_aAnswers.empty())
			return 0;

		//Get the answer
		const CDNSAnswer& rAnswer=m_aAnswers[0];

		//Check what type is it ?
		if (rAnswer.GetResourceType()==CDNSQuery::A)
		{
			//Get the data
			std::string sData;
			sData=rAnswer.GetData();

			//Convert it to long
			return atol(sData.c_str());
		}
		else if (rAnswer.GetResourceType()==CDNSQuery::MX &&
				 m_aDNSHeader.usARCount)
		{
			//Look for the additional resources
			int iPos;
			iPos=m_aDNSHeader.usANCount+
				 m_aDNSHeader.usNSCount;

			//Our data
			std::string sData;
			sData=rAnswer.GetData();

			//Loop them
			for (int iCount=0;
				 iCount<m_aDNSHeader.usARCount;
				 iCount++)
			{
				//Get the answer
				const CDNSAnswer& rARAnswer=m_aAnswers[iCount+iPos];

				//Is it an A record?
				if (rARAnswer.GetResourceType()==CDNSQuery::A &&
					rARAnswer.GetName()==sData)
					//This is us
					return atol(rARAnswer.GetData().c_str());
			}

			//Not found
			return 0;
		}
		else
			return 0;
	}
	ERROR_HANDLER_RETURN("GetAIPAnswer",0)
}

CDNSAnswers::IPAnswersVector CDNSAnswers::GetAIPAnswers()const
{
	//Empty vector
	static IPAnswersVector aEmptyVector;

	try
	{
		//Our vector
		IPAnswersVector aVector;

		//Do we have an answer ?
		if (!m_aDNSHeader.usANCount ||
			m_aAnswers.empty())
			return aVector;

		//Iterate the answers
		for (int iCount=0;
			 iCount<m_aDNSHeader.usANCount;
			 ++iCount)
		{
			//Get the answer
			const CDNSAnswer& rAnswer=m_aAnswers[iCount];

			//Check what type is it ?
			if (rAnswer.GetResourceType()==CDNSQuery::A)
			{
				//Get the data
				std::string sData;
				sData=rAnswer.GetData();

				//Convert it to long
				aVector.push_back(atol(sData.c_str()));
			}
			else if (rAnswer.GetResourceType()==CDNSQuery::MX &&
					 m_aDNSHeader.usARCount)
			{
				//Look for the additional resources
				int iPos;
				iPos=m_aDNSHeader.usANCount+
					 m_aDNSHeader.usNSCount;

				//Our data
				std::string sData;
				sData=rAnswer.GetData();

				//Loop them
				for (int iCount=0;
					 iCount<m_aDNSHeader.usARCount;
					 iCount++)
				{
					//Get the answer
					const CDNSAnswer& rARAnswer=m_aAnswers[iCount+iPos];

					//Is it an A record? (may have more then one A record per type)
					if (rARAnswer.GetResourceType()==CDNSQuery::A &&
						rARAnswer.GetName()==sData)
						//This is us
						aVector.push_back(atol(rARAnswer.GetData().c_str()));
				}
			}
		}

		//Done
		return aVector;
	}
	ERROR_HANDLER_RETURN("GetAIPAnswers",aEmptyVector)
}

std::string CDNSAnswers::GetAnswer()const
{
	try
	{
		//Do we have an answer ?
		if (!m_aDNSHeader.usANCount ||
			m_aAnswers.empty())
			return "";

		//Get the answer
		const CDNSAnswer& rAnswer=m_aAnswers[0];

		//Check what type is it ?
		if (rAnswer.GetResourceType()==CDNSQuery::PTR ||
			rAnswer.GetResourceType()==CDNSQuery::MX)
		{
			//Get the data
			std::string sData;
			sData=rAnswer.GetData();

			//Convert it to long
			return sData;
		}
		else if (rAnswer.GetResourceType()==CDNSQuery::A)
		{
			//Get the data
			std::string sData;
			sData=rAnswer.GetData();

			//Convert to DWORD
			DWORD dwIP;
			dwIP=atof(sData.c_str());

			//Convert to string
			return CSpoofBase::LongToStdString(dwIP);
		}
		else
			return "";
	}
	ERROR_HANDLER_RETURN("GetAnswer","")
}

CDNSAnswers::StringAnswersVector CDNSAnswers::GetAnswers()const
{
	//Empty vector
	static StringAnswersVector aEmptyVector;

	try
	{
		//Our vector
		StringAnswersVector aVector;

		//Do we have an answer ?
		if (!m_aDNSHeader.usANCount ||
			m_aAnswers.empty())
			return aEmptyVector;

		//Iterate the answers
		for (int iCount=0;
			 iCount<m_aDNSHeader.usANCount;
			 ++iCount)
		{
			//Get the answer
			const CDNSAnswer& rAnswer=m_aAnswers[iCount];

			//Check what type is it ?
			if (rAnswer.GetResourceType()==CDNSQuery::PTR ||
				rAnswer.GetResourceType()==CDNSQuery::MX)
			{
				//Get the data
				std::string sData;
				sData=rAnswer.GetData();

				//Add it
				aVector.push_back(sData);
			}
			else if (rAnswer.GetResourceType()==CDNSQuery::A)
			{
				//Get the data
				std::string sData;
				sData=rAnswer.GetData();

				//Convert to DWORD
				DWORD dwIP;
				dwIP=atof(sData.c_str());

				//Convert to string
				//Add it
				aVector.push_back(CSpoofBase::LongToStdString(dwIP));
			}
		}

		//Done
		return aVector;
	}
	ERROR_HANDLER_RETURN("GetAnswers",aEmptyVector)
}

const CDNSQuery* CDNSAnswers::GetQuery()const
{
	return m_pDNSQuery;
}

std::string CDNSAnswers::GetAnswerParsed(BOOL bWindowsLines,
										 BOOL bBrief,
										 BOOL bExtraLine)const
{
	try
	{
		//Line terminator
		//End of line
		char aEndOfLine[3];

		//Which is it
		if (bWindowsLines)
		{
			aEndOfLine[0]=13;
			aEndOfLine[1]=10;
			aEndOfLine[2]=0;
		}
		else
		{
			aEndOfLine[0]='\n';
			aEndOfLine[1]=0;
		}

		//Our string
		std::string sAnswer;

		//Iterate answers
		for (int iCount=0;
			 iCount<m_aAnswers.size();
			 iCount++)
		{
			//What are we
			if (iCount<m_aDNSHeader.usANCount &&
				!iCount)
			{
				//Answer
				sAnswer+="Non-authoritative answer:";
				sAnswer+=aEndOfLine;
			}
			else if ((iCount<m_aDNSHeader.usANCount+
							 m_aDNSHeader.usNSCount) &&
					 iCount==m_aDNSHeader.usANCount)
			{
				//Authority servers
				sAnswer+=aEndOfLine;
				sAnswer+="Authoritive servers:";
				sAnswer+=aEndOfLine;
			}
			else if (iCount==m_aDNSHeader.usANCount+
							 m_aDNSHeader.usNSCount)
			{
				//Additional records
				sAnswer+=aEndOfLine;
				sAnswer+="Additional records:";
				sAnswer+=aEndOfLine;
			}

			//Add the answer
			sAnswer+=m_aAnswers[iCount].GetAnswerParsed(bWindowsLines,
														bBrief);
			sAnswer+=aEndOfLine;
		}

		//Do we need another line?
		if (bExtraLine)
			sAnswer+=aEndOfLine;

		//Done
		return sAnswer;
	}
	ERROR_HANDLER_RETURN("GetAnswerParsed","")
}

BOOL CDNSAnswers::IsReverseAnswer()const
{
	return IsResolved() &&
		   m_pDNSQuery->GetFirstQueryType()==CDNSQuery::PTR;
}

//of namespace
}