/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "DNSManager.h"

#include "ErrorHandlerMacros.h"
#include "Array_ptr.h"
#include "OSManager.h"
#include "GenericCriticalSection.h"
#include "GenericThread.h"
#include "ThreadPool.h"

#include "DNSSocket.h"
#include "DNSQuery.h"

#include <memory>

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

namespace KomodiaDNS
{

#define CDNSManager_Class "CDNSManager"

#define DELETE_TIME 120000

CDNSManager::CDNSManager(const std::string strDNSServer,
						 BOOL bTCP,
						 BOOL bAsync,
						 BOOL bAsyncConnect) : CErrorHandler(),
											   m_pSocket(NULL),
											   m_strDNSServer(strDNSServer),
											   m_iParseThreads(0),
											   m_pCSection(NULL),
											   m_bInitialized(FALSE),
											   m_pThread(NULL),
											   m_dwTimeout(0)
{
	try
	{
		//Set our name
		SetName(CDNSManager_Class);

		//Set the values
		m_bAsync=bAsync;
		m_bTCP=bTCP;

		//Create the CS
		m_pCSection=COSManager::CreateCriticalSection();

		//Allocate our manager
		m_pSocket=new CDNSSocket(this,
								m_strDNSServer,
								m_bTCP,
								m_bAsync,
								bAsyncConnect);

		//Are we async ?
		if (m_bAsync)
		{
			//Create the timeout thread
			m_pThread=new CPeriodicThread(ThreadProc);
			
			//Set it's priority
			m_pThread->SetPriority(CGenericThread::tpAboveNormal);

			//Start it
			if (!m_pThread->Start(500,
								  (LPVOID)this))
				//Report it
				ReportError("CDNSManager","Failed to start periodic thread!");
		}
	}
	ERROR_HANDLER("CDNSManager")
}

CDNSManager::~CDNSManager()
{
	try
	{
		//Delete the socket
		delete m_pSocket;

		//Delete the thread
		delete m_pThread;

		//And the CS
		delete m_pCSection;
	}
	ERROR_HANDLER("~CDNSManager")
}

CThreadPool* CDNSManager::GetThreadManager()const
{
	return m_pSocket->GetThreadManager();
}

std::string CDNSManager::ParseAddress(const std::string& rAddress)
{
	try
	{
		//Our address
		std::string sAddress;

		//Index count
		int iIndex=0;

		//Our last found position
		std::string::size_type iLastPosition;
		iLastPosition=0;

		//Our current position
		std::string::size_type iFoundPos;

		//Start parsing
		while ((iFoundPos=rAddress.find_first_of('.',iLastPosition))!=std::string::npos)
		{
			//Put the position in the new string
			sAddress+=(char)(iFoundPos-iLastPosition);

			//Add the partial string
			sAddress+=rAddress.substr(iLastPosition,
									  iFoundPos-iLastPosition);

			//Set the last position to be the current
			iLastPosition=iFoundPos+1;
		}

		//Only if not null
		if (rAddress.length()!=iLastPosition)
		{
			//Insert the last part of the string
			sAddress+=(char)(rAddress.length()-iLastPosition);
			sAddress+=rAddress.substr(iLastPosition,
									  rAddress.length()-iLastPosition);
		}

		//Done
		return sAddress;
	}
	ERROR_HANDLER_STATIC_RETURN(CDNSManager_Class,"ParseAddress",NULL)
}

LPDnsHeaderHeader CDNSManager::GetDNSHeaderHeader(unsigned short usID,
												  unsigned short usFlags,
												  unsigned short usQuestions,
												  unsigned short usAnswers)const
{
	try
	{
		//Build the header
		LPDnsHeaderHeader lpHead;
		lpHead=new DnsHeaderHeader;

		lpHead->usID=htons(usID);
		lpHead->usOptions=usFlags;
		lpHead->usQDCount=htons(usQuestions);
		lpHead->usANCount=htons(usAnswers);

		//Done
		return lpHead;
	}
	ERROR_HANDLER_RETURN("GetDNSHeaderHeader",NULL)
}

unsigned short CDNSManager::QueryDNS(const CDNSQuery& rQuery,
									 unsigned short usID,
									 LPVOID lpLocalIDData)
{
	try
	{
		//Are we initialized
		if (!m_bInitialized)
		{
			//Report it
			ReportError("QueryDNS","Not initialized!");
			
			//Exit
			return 0;
		}

		//Get the query size
		unsigned short usTotalSize;
		usTotalSize=rQuery.GetSerializationSize();

		//Is it bad?
		if (!usTotalSize)
		{
			//Report it
			ReportError("QueryDNS","Bad query size!");
			
			//Exit
			return 0;
		}

		//Add the dns header
		usTotalSize+=DnsHeaderHeaderLength;

		//Is it TCP
		if (m_bTCP)
			usTotalSize+=2;

		//Allocate the buffer
		char* pBuffer;
		pBuffer=new char[usTotalSize];

		//Protect it
		CArray_ptr<char> pProtection(pBuffer);

		//Keep track of position
		char* pBackupBuffer=pBuffer;

		if (m_bTCP)
			//Advance pointer
			pBuffer+=2;

		//Get the ID
		if (!usID)
		{
			//Get a running static ID
			static long lID=0;
			InterlockedIncrement(&lID);
			usID=lID;
		}

		//Get the DNS header 
		LPDnsHeaderHeader lpHead;
		lpHead=GetDNSHeaderHeader(usID,
								  DNS_RECURSION,
								  rQuery.GetQuestionCount());

		//Copy it
		memcpy(pBuffer,
			   lpHead,
			   DnsHeaderHeaderLength);
		pBuffer+=DnsHeaderHeaderLength;

		//Done with it
		delete lpHead;

		//Serialize the query
		if (!rQuery.SerializeQuery(pBuffer))
		{
			//Report it
			ReportError("QueryDNS","Failed to serialize query!");
			
			//Exit
			return 0;
		}

		//Fix the tcp header
		if (m_bTCP)
			*((unsigned short*)pBackupBuffer)=htons(usTotalSize-2);

		//Add the record
		AddRecord(usID,
				  lpLocalIDData,
				  m_dwTimeout,
				  rQuery);

		//Send the data
		BOOL bResult;
		bResult=m_pSocket->Send(pBackupBuffer,
							    usTotalSize,
							    lpLocalIDData);

		//If error remove the data
		if (!bResult)
			RemoveRecord(usID);

		//Check if all is OK
		if (bResult)
			return usID;
		else
			return 0;
	}
	ERROR_HANDLER_RETURN("QueryDNS",0)
}

unsigned short  CDNSManager::GetDNSEntry(const std::string& rAddress,
										 unsigned short usID,
										 LPVOID lpLocalIDData)
{
	try
	{
		//Convert it to long
		unsigned long ulAddress;
		ulAddress=CSpoofBase::StringToLong(rAddress);

		//And run the query
		return GetDNSEntry(ulAddress,
						   usID,
						   lpLocalIDData);
	}
	ERROR_HANDLER_RETURN("GetDNSEntry",0)
}

unsigned short CDNSManager::GetDNSEntry(unsigned long ulAddress,
										unsigned short usID,
										LPVOID lpLocalIDData)
{
	try
	{
		//Reverse the address
		long ulNewAddress;
		ulNewAddress=htonl(ulAddress);

		//Convert it to string
		std::string sAddr;
		sAddr=CSpoofBase::LongToStdString(ulNewAddress);

		//Add the arpa stuff
		sAddr+=".in-addr.arpa.";

		//Create the query
		CDNSQuery aQuery;
		aQuery.AddQuery(sAddr.c_str(),
						CDNSQuery::PTR);

		//and query it
		return QueryDNS(aQuery,
						usID,
						lpLocalIDData);
	}
	ERROR_HANDLER_RETURN("GetDNSEntry",0)
}

void CDNSManager::OnDNSReceive(CDNSAnswers* pAnswers,
							   LPVOID lpLocalIDData)
{
	try
	{
		//Here we received an answer, since a client should inherit this, I will delete it, to avoid leaks
		delete pAnswers;
	}
	ERROR_HANDLER("OnDNSReceive")
}

void CDNSManager::OnDNSTimeout(const CDNSQuery& rQuery,
							   LPVOID lpLocalIDData)
{
}

void CDNSManager::OnDNSError(int iErrorCode,
							 LPVOID lpLocalIDData)
{
}

void CDNSManager::ParseMultithreaded(int iThreadNumber)
{
	try
	{
		//Set the data
		m_iParseThreads=iThreadNumber;

		//And the socket
		m_pSocket->SetMultithreaded(m_iParseThreads);
	}
	ERROR_HANDLER("ParseMultithreaded")
}

BOOL CDNSManager::Initialize()
{
	try
	{
		if (m_bInitialized)
			return TRUE;
		else
		{
			//Initialize the sockets
			m_bInitialized=m_pSocket->Initialize();

			//Exit
			return m_bInitialized;
		}
	}
	ERROR_HANDLER_RETURN("Initialize",FALSE)
}

BOOL CDNSManager::IsInitialized()const
{
	return m_bInitialized;
}

void CDNSManager::Uninitialize()
{
	try
	{
		if (!m_bInitialized)
			return;
		else
		{
			//Not initialized
			m_bInitialized=FALSE;
			
			//Delete the socket
			delete m_pSocket;
			m_pSocket=NULL;
		}
	}
	ERROR_HANDLER("Uninitialize")
}

void CDNSManager::OnDNSReceive(CDNSAnswers* pAnswers)
{
	try
	{
		//Protect the answers
		std::auto_ptr<CDNSAnswers> pProtection(pAnswers);

		//Get the ID
		unsigned short usID;
		usID=pAnswers->GetDNSID();

		//Where we timed out ?
		if (WasTimedout(usID))
			return;

		//Get the data
		LPVOID lpData;
		lpData=GetRecord(usID);

		//Release the protection
		pProtection.release();

		//Fire the event
		OnDNSReceive(pAnswers,lpData);

		//Delete the record
		RemoveRecord(usID);
	}
	ERROR_HANDLER("OnDNSReceive")
}

CDNSAnswers* CDNSManager::Receive()
{
	try
	{
		//Are we initialized
		if (!m_bInitialized)
		{
			//Report it
			ReportError("Receive","Not initialized!");
			
			//Exit
			return NULL;
		}

		//Call socket
		return m_pSocket->Receive();
	}
	ERROR_HANDLER_RETURN("Receive",NULL)
}

void CDNSManager::AddRecord(unsigned short usID,
							LPVOID lpLocalIDData,
							DWORD dwTimeout,
							const CDNSQuery& rQuery)
{
	try
	{
		//Create the struct
		DNSData aData;
		aData.aQuery=rQuery;
		aData.dwTime=GetTickCount();
		aData.lpData=lpLocalIDData;
		aData.dwTimeout=dwTimeout;
		aData.bTimedout=FALSE;
		aData.bNotified=FALSE;
		aData.dwTimedoutTime=0;

		//Lock the data
		CCriticalAutoRelease aRelease(m_pCSection);

		//Delete the record if exists
		m_aData.erase(usID);

		//Insert the data
		m_aData.insert(IDMap::value_type(usID,aData));
	}
	ERROR_HANDLER("AddRecord")
}

BOOL CDNSManager::WasTimedout(unsigned short usID)
{
	try
	{
		//Lock the data
		CCriticalAutoRelease aRelease(m_pCSection);

		//Try to find it
		IDMap::iterator aIterator;
		aIterator=m_aData.find(usID);

		//Do we have it
		if (aIterator!=m_aData.end())
			if (aIterator->second.bTimedout)
				return TRUE;
			else
			{
				//Set it to be notified
				aIterator->second.bNotified=TRUE;

				//Exit
				return FALSE;
			}
		else
			return FALSE;
	}
	ERROR_HANDLER_RETURN("WasTimedout",TRUE)
}

LPVOID CDNSManager::GetRecord(unsigned short usID)const
{
	try
	{
		//Lock the data
		CCriticalAutoRelease aRelease(m_pCSection);

		//Try to find it
		IDMap::const_iterator aIterator;
		aIterator=m_aData.find(usID);

		//Do we have it
		if (aIterator!=m_aData.end())
			return aIterator->second.lpData;
		else
			return NULL;
	}
	ERROR_HANDLER_RETURN("GetRecord",NULL)
}

void CDNSManager::RemoveRecord(unsigned short usID)
{
	try
	{
		//Lock the data
		CCriticalAutoRelease aRelease(m_pCSection);

		//Delete it
		m_aData.erase(usID);
	}
	ERROR_HANDLER("RemoveRecord")
}

void CDNSManager::SetDNSTimeout(DWORD dwMS)
{
	try
	{
		//Are we initialized
		if (!m_bInitialized)
		{
			//Report it
			ReportError("SetTimeout","Not initialized!");
			
			//Exit
			return;
		}

		//Save the new timeout
		m_dwTimeout=dwMS;
	}
	ERROR_HANDLER("SetDNSTimeout")
}

BOOL CDNSManager::SetConnectionTimeout(DWORD dwMS)
{
	try
	{
		//Are we initialized
		if (!m_bInitialized)
		{
			//Report it
			ReportError("SetTimeout","Not initialized!");
			
			//Exit
			return NULL;
		}

		//Set it
		return m_pSocket->SetConnectionTimeout(dwMS);
	}
	ERROR_HANDLER_RETURN("SetTimeout",FALSE)
}

BOOL CDNSManager::ThreadProc(CPeriodicThread::ThreadStage aStage,
	  					     LPVOID lpData)
{
	try
	{
		//Only if we are running
		if (aStage!=CPeriodicThread::tsBody)
			return TRUE;

		//Get our class
		CDNSManager* pClass;
		pClass=(CDNSManager*)lpData;

		//Lock the CCsection
		CCriticalAutoRelease aRelease(pClass->m_pCSection);

		//Scan the data
		IDMap::iterator aIterator;
		aIterator=pClass->m_aData.begin();

		//Our timeout vector
		typedef std::vector<DNSData> TimeoutVector;
		TimeoutVector aTimeoutVector;

		//Iterate it now
		while (aIterator!=pClass->m_aData.end())
		{
			//Check the times
			if (aIterator->second.dwTimeout && 
				GetTickCount()-aIterator->second.dwTime>aIterator->second.dwTimeout &&
				!aIterator->second.bNotified &&
				!aIterator->second.bTimedout)
			{
				//It's a timeout
				//Set ir
				aIterator->second.bTimedout=TRUE;
				aIterator->second.dwTimedoutTime=GetTickCount();

				//Insert it
				aTimeoutVector.push_back(aIterator->second);

				//Next data
				++aIterator;	
			}
			else if (GetTickCount()-aIterator->second.dwTimedoutTime>DELETE_TIME &&
					 aIterator->second.bTimedout)
			{
				//Save the data
				IDMap::iterator aBackupIterator=aIterator;
				++aIterator;

				//And delete it
				pClass->m_aData.erase(aBackupIterator);
			}
			else
				//Next data
				++aIterator;
		}

		//Exit the CS
		aRelease.Exit();

		//Spawn the timeouts
		TimeoutVector::const_iterator aTimeoutIterator;
		aTimeoutIterator=aTimeoutVector.begin();

		//Iterate it
		while (aTimeoutIterator!=aTimeoutVector.end())
		{
			//Spawn a timeout
			pClass->OnDNSTimeout(aTimeoutIterator->aQuery,
								 aTimeoutIterator->lpData);

			//Next item
			++aTimeoutIterator;
		}

		//Done
		return TRUE;
	}
	ERROR_HANDLER_STATIC_RETURN(CDNSManager_Class,"ThreadProc",THREAD_DO_NOTHING_EXIT_VALUE)
}

void CDNSManager::ParseMultithreaded(CThreadPool* pThreadManager)
{
	try
	{
		//Pass it to the socket
		m_pSocket->SetMultithreaded(pThreadManager);
	}
	ERROR_HANDLER("ParseMultithreaded")
}

//of namespace
}