/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#if !defined(AFX_DNSMANAGER_H__4B050500_D65B_4CA1_A801_CA06418204E5__INCLUDED_)
#define AFX_DNSMANAGER_H__4B050500_D65B_4CA1_A801_CA06418204E5__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#pragma warning(disable:4786)

#include "ErrorHandler.h"
#include "PeriodicThread.h"

#include "DNSAnswers.h"

#include <string>
#include <map>

class CGenericCriticalSection;
class CThreadPool;

namespace KomodiaDNS
{

//Type of headers
typedef enum _DnsHeaderType
{
	dhQuestion,
	dhAnswer,
	dhAuthoritive,
	dhAdditional,
	dhNone
} DnsHeaderType;

//DNS System type
#define DNS_CLASS_INTERNET 1
#define DNS_CLASS_CSNET 2
#define DNS_CLASS_CHAOS 3
#define DNS_CLASS_HESIOD 4

class CDNSSocket;
class CDNSQuery;

class CDNSManager : public CErrorHandler
{
public:
	friend class CDNSSocket;
public:
	//Get the thread manager
	CThreadPool* GetThreadManager()const;

	//Set the timeout for the DNS
	void SetDNSTimeout(DWORD dwMS);

	//Set the timeout (for the sockets)
	BOOL SetConnectionTimeout(DWORD dwMS);

	//Clean the data (incase we are static)
	void Uninitialize();

	//Receive data
	CDNSAnswers* Receive();

	//Initialize ourselves
	BOOL Initialize();

	//Are we initialized
	BOOL IsInitialized()const;

	//Get the dns entry of an address
	//usID - ID of the DNS (if zero, library will allocate one)
	//lpLocalIDData - Custome user data
	unsigned short GetDNSEntry(const std::string& rAddress,
							   unsigned short usID=0,
							   LPVOID lpLocalIDData=NULL);
	unsigned short GetDNSEntry(unsigned long ulAddress,
							   unsigned short usID=0,
							   LPVOID lpLocalIDData=NULL);

	//Allow to parse multithreaded
	void ParseMultithreaded(int iThreadNumber);
	void ParseMultithreaded(CThreadPool* pThreadManager);

	//Send a query (Returns the Query ID)
	unsigned short QueryDNS(const CDNSQuery& rQuery,
							unsigned short usID=0,
							LPVOID lpLocalIDData=NULL);

	//Parse an address, convert it to DNS
	static std::string ParseAddress(const std::string& rAddress);

	//Ctor and Dtor
	//bTCP - Use DNS TCP
	//bAsync - Use async sockets
	//bAsyncConnect - Connect blockingly
	CDNSManager(const std::string strDNSServer,
				BOOL bTCP,
				BOOL bAsync,
				BOOL bAsyncConnect=FALSE);
	virtual ~CDNSManager();
protected:
	//Get a generic DNS
	virtual LPDnsHeaderHeader GetDNSHeaderHeader(unsigned short usID,
												 unsigned short usFlags,
												 unsigned short usQuestions,
												 unsigned short usAnswers=0)const;

	//Events for user to override
	virtual void OnDNSError(int iErrorCode,
							LPVOID lpLocalIDData);
	virtual void OnDNSTimeout(const CDNSQuery& rQuery,
							  LPVOID lpLocalIDData);
	virtual void OnDNSReceive(CDNSAnswers* pAnswers,
							  LPVOID lpLocalIDData);
private:
	//Data inside the map
	typedef struct _DNSData
	{
		CDNSQuery	aQuery;
		LPVOID		lpData;
		DWORD		dwTime;
		DWORD		dwTimedoutTime;
		DWORD		dwTimeout;
		BOOL		bTimedout;
		BOOL		bNotified;
	} DNSData;

	//Our map of ID data
	typedef std::map<unsigned short,DNSData> IDMap;
private:
	//No copy Ctor
	CDNSManager(const CDNSManager& rManager);

	//No copy operator
	CDNSManager& operator=(const CDNSManager& rManager);

	//Receive an answer from the sockets
	void OnDNSReceive(CDNSAnswers* pAnswers);

	//Add a record
	void AddRecord(unsigned short usID,
				   LPVOID lpLocalIDData,
				   DWORD dwTimeout,
				   const CDNSQuery& rQuery);

	//Get a record
	LPVOID GetRecord(unsigned short usID)const;

	//Remove a record
	void RemoveRecord(unsigned short usID);

	//Did we timeout
	BOOL WasTimedout(unsigned short usID);

	//Thread proc to iterate the records
	static BOOL ThreadProc(CPeriodicThread::ThreadStage aStage,
						   LPVOID lpData);

	//Our query data
	IDMap m_aData;

	//Status flags
	BOOL m_bAsync;
	BOOL m_bTCP;

	//Our socket
	CDNSSocket* m_pSocket;

	//Our DNS server
	std::string m_strDNSServer;

	//Do we parse multithreaded
	int m_iParseThreads;

	//Our cs
	CGenericCriticalSection* m_pCSection;

	//Are we initialized
	BOOL m_bInitialized;

	//Our periodic thread
	CPeriodicThread* m_pThread;

	//The timeout
	DWORD m_dwTimeout;
};

//of namespace
}

#endif // !defined(AFX_DNSMANAGER_H__4B050500_D65B_4CA1_A801_CA06418204E5__INCLUDED_)
