/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "TCPPortScannerStealth.h"

#include "ErrorHandlerMacros.h"
#include "GenericCriticalSection.h"
#include "OSManager.h"

#include <vector>

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

KOMODIA_NAMESPACE_START

// ----------------------------- CSniffingSocket start ---------------------

#define CSniffingSocket_Class "CSniffingSocket"

CTCPPortScannerStealth::CSniffingSocket::CSniffingSocket(CTCPPortScannerStealth* pScanner) : CSniffSocket(),
																							 m_bStop(FALSE),
																							 m_pFather(pScanner)
{
	try
	{
		//Set our name
		SetName(CSniffingSocket_Class);
	}
	ERROR_HANDLER("CSniffingSocket")
}

CTCPPortScannerStealth::CSniffingSocket::~CSniffingSocket()
{
}

BOOL CTCPPortScannerStealth::CSniffingSocket::OnSocketReceive(int iErrorCode)
{
	try
	{
		char cBuffer[2000];

		//First receive the packet
		int iReceive;
		iReceive=Receive(cBuffer,
						 sizeof(cBuffer));

		//Are we stopped
		if (m_bStop)
			return FALSE;

		if (iReceive==GetErrorCode())
			return FALSE;

		//Copy the header
		IpHeader aHeader;
		memcpy(&aHeader,
			   cBuffer,
			   IpHeaderLength);

		//Check the data is not from us (altough someone may want to save this data)
		if (aHeader.ucProtocol==IPPROTO_TCP && 
			aHeader.ulDestinationAddress==m_pFather->GetBindInterface() && 
			aHeader.ulSourceAddress==m_pFather->GetTarget())
		{
			//Find the size of IP header (may have options)
			unsigned long ulHeaderSize;
			ulHeaderSize=(aHeader.ucHeaderLength_Version & 0x0f) << 2;

			//And check is it the scan ?
			AnalyzeTCP(cBuffer+ulHeaderSize);
		}

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("OnSocketReceive",TRUE)
}

void CTCPPortScannerStealth::CSniffingSocket::AnalyzeTCP(const char* pTCPBuffer)
{
	try
	{
		//Read the ICMP header
		TCPHeader aHeader;
		memcpy(&aHeader,
			   pTCPBuffer,
			   TCPHeaderLength);

		//Our port
		unsigned short usPort;
		usPort=htons(aHeader.usSourcePort);

		//Is it connection
		int iConnection;
		iConnection=-1;

		//Check the flag
		if (aHeader.ucFlags==(TCPFlag_SYN | TCPFlag_ACK))
			//Connected
			iConnection=1;
		else if (aHeader.ucFlags==(TCPFlag_RST | TCPFlag_ACK))
			//Not connected
			iConnection=0;

		//Are we connected
		if (iConnection!=-1)
		{
			//Remove from list
			m_pFather->RemovePort(usPort);

			//Indicate status
			m_pFather->SocketDone(usPort,
								  iConnection);
		}
	}
	ERROR_HANDLER("AnalyzeTCP")
}

void CTCPPortScannerStealth::CSniffingSocket::Stop()
{
	m_bStop=TRUE;
}

// ----------------------------- CSniffingSocket end ---------------------

#define CTCPPortScannerStealth_Class "CTCPPortScannerStealth"

//Scan timeout
#define SCAN_TIMEOUT 5000

CTCPPortScannerStealth::CTCPPortScannerStealth(IP aBindInterface) : CTCPPortScanner(),
																	m_pScanner(NULL),
																	m_aBindInterface(aBindInterface),
																	m_pSniffer(NULL),
																	m_pCSection(NULL),
																	m_pThread(NULL)
{
	try
	{
		//Set our name
		SetName(CTCPPortScannerStealth_Class);

		//Create the CS
		m_pCSection=COSManager::CreateCriticalSection();
	}
	ERROR_HANDLER("CTCPPortScannerStealth")
}

CTCPPortScannerStealth::CTCPPortScannerStealth(const std::string& rBindInterface) : CTCPPortScanner(),
																					m_pScanner(NULL),
																					m_aBindInterface(0),
																					m_pSniffer(NULL),
																					m_pCSection(NULL),
																					m_pThread(NULL)
{
	try
	{
		//Set our name
		SetName(CTCPPortScannerStealth_Class);

		//Create the CS
		m_pCSection=COSManager::CreateCriticalSection();

		//Convert the bind interface
		m_aBindInterface=CSpoofBase::StringToLong(rBindInterface);
	}
	ERROR_HANDLER("CTCPPortScannerStealth")
}

CTCPPortScannerStealth::~CTCPPortScannerStealth()
{
	try
	{
		//Delete the thread
		delete m_pThread;

		//Delete the scanner
		delete m_pScanner;

		//Delete the sniffer
		delete m_pSniffer;

		//Delete the CS
		delete m_pCSection;
	}
	ERROR_HANDLER("~CTCPPortScannerStealth")
}

BOOL CTCPPortScannerStealth::CreateSniffer()
{
	try
	{
		//Try to create the sniffer socket
		if (m_pSniffer)
		{
			delete m_pSniffer;
			m_pSniffer=NULL;
		}

		//Create it	
		CSniffingSocket* pSocket;
		pSocket=new CSniffingSocket(this);

		//Protect it
		std::auto_ptr<CSniffingSocket> pProtection(pSocket);

		//Continue to create it
		if (!pSocket->Create())
		{
			//Report it
			ReportError("CreateSniffer","Failed to create sniffer");

			//Exit
			return FALSE;
		}
		
		//Try to bind it
		if (!pSocket->Bind(m_aBindInterface,0))
		{
			//Report it
			ReportError("CreateSniffer","Failed to bind sniffer");

			//Exit
			return FALSE;
		}

		//Release protection and save socket
		m_pSniffer=pProtection.release();

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("CreateSniffer",FALSE)
}

BOOL CTCPPortScannerStealth::Scan(const std::string& rDestinationAddress)
{
	try
	{
		return Scan(CSpoofBase::StringToLong(rDestinationAddress));
	}
	ERROR_HANDLER_RETURN("Scan",FALSE)
}

BOOL CTCPPortScannerStealth::Scan(IP aTarget)
{
	//Check max sockets are ok
	if (GetMaxSockets()<=0 ||
		!IsDone() ||
		!aTarget)
	{
		//Report it
		ReportError("Scan","Invalid parameters!");

		//Exit
		return FALSE;
	}

	try
	{
		//Delete the thread
		delete m_pThread;
		m_pThread=NULL;

		//Reset the pair
		ResetPair();

		//Can we scan
		if (!CanScan())
		{
			//Error
			ReportError("Scan","No ports given!");

			//Error
			SetError(TRUE);

			//Exit
			return FALSE;
		}

		//Clear the ports map
		m_aPortsMap.clear();

		//Do we have the socket?
		if (m_pScanner)
			delete m_pScanner;

		//Allocate it
		m_pScanner=new CTCPScanner;

		//Try to create it
		if (!m_pScanner->Create())
		{
			//Report it
			ReportError("Scan","Failed to create socket!");

			//Exit
			return FALSE;
		}

		//Set the source address as the sniffing interface
		m_pScanner->SetSourceAddress(m_aBindInterface);

		//Try to create the sniffer
		if (!CreateSniffer())
		{
			//Error
			ReportError("Scan","Failed to create sniffer!");

			//Error
			SetError(TRUE);

			//Exit
			return FALSE;
		}

		//Set the flags
		SetError(FALSE);
		SetDone(FALSE);

		//Save the number of max sockets
		ResetSocketCount();

		//Delete the port data
		DeletePortsMap();

		//Delete other data
		ResetData();

		//Set our target
		SetTarget(aTarget);

		//Start the sniffer
		if (!m_pSniffer->Sniff(TRUE))
		{
			//Error
			ReportError("Scan","Failed to create sniffer!");

			//Error
			SetError(TRUE);
			
			//We are done
			SetDone(TRUE);

			//Exit
			return FALSE;
		}

		//Start the thread
		StartThread();

		//Start the loop
		for (int iCounter=0;
			 iCounter<GetMaxSockets();
			 ++iCounter)
		{
			BOOL bQuit;
			bQuit=FALSE;

			//Try to allocate sockets
			while (!bQuit)
			{
				//Another socket
				AdjustNumberOfSockets(1);

				//Create it
				if (!(bQuit=NewSocket()))
					ReportError("Scan","Failed to create socket!");
			}
		}

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("Scan",FALSE)
}

CTCPSocketAsync* CTCPPortScannerStealth::AllocateSocket(unsigned short usPort)const
{
	return m_pScanner;
}

void CTCPPortScannerStealth::DestroySocket(CTCPSocketAsync* pSocket)const
{
	//Can't delete our socket!
}

IP CTCPPortScannerStealth::GetBindInterface()const
{
	return m_aBindInterface;
}

void CTCPPortScannerStealth::TCPScanDoneLib(BOOL bError)
{
	try
	{
		//Do we have a sniffer?
		if (m_pSniffer)
		{
			//Stop the sniffer
			m_pSniffer->Stop();

			//Delete it from it's thread
			m_pSniffer->DeleteSocketFromThread();

			//Remove it
			m_pSniffer=NULL;
		}

		//Delete the thread
		delete m_pThread;
		m_pThread=NULL;
	}
	ERROR_HANDLER("TCPScanDoneLib")
}

void CTCPPortScannerStealth::PortScanned(unsigned short usPort)
{
	try
	{
		//Lock
		CCriticalAutoRelease aRelease(m_pCSection);

		//Add an entry to the map
		m_aPortsMap.insert(PortsMap::value_type(usPort,GetTickCount()));
	}
	ERROR_HANDLER("PortScanned")
}

void CTCPPortScannerStealth::RemovePort(unsigned short usPort)
{
	try
	{
		//Lock
		CCriticalAutoRelease aRelease(m_pCSection);

		//Add an entry to the map
		m_aPortsMap.erase(usPort);
	}
	ERROR_HANDLER("RemovePort")
}

BOOL CTCPPortScannerStealth::TimeoutProc(CPeriodicThread::ThreadStage aStage,
										 LPVOID pParam)
{
	try
	{
		//Which stage are we
		if (aStage!=CPeriodicThread::tsBody)
			return TRUE;

		//The class
		CTCPPortScannerStealth* pClass;
		pClass=(CTCPPortScannerStealth*)pParam;

		//Which timeout do we use
		unsigned long ulTimeout;
		if (!(ulTimeout=pClass->GetConnectionTimeout()))
			ulTimeout=SCAN_TIMEOUT;

		//Our vector of ports
		typedef std::vector<unsigned short> PortsVector;
		PortsVector aVector;

		{
			//Our tick count
			DWORD dwTick;
			dwTick=GetTickCount();

			//Lock
			CCriticalAutoRelease aRelease(pClass->m_pCSection);

			//Iterate the data
			PortsMap::iterator aIterator;
			aIterator=pClass->m_aPortsMap.begin();
			while (aIterator!=pClass->m_aPortsMap.end() &&
				   !pClass->IsFinished())
			{
				//Check if passed it's time
				if (dwTick-aIterator->second>=ulTimeout)
				{
					//Add to the vector
					aVector.push_back(aIterator->first);

					//Erase the data
					aIterator=pClass->m_aPortsMap.erase(aIterator);
				}
				else
					//Next
					++aIterator;
			}
		}

		//Start to shoot ports events
		PortsVector::const_iterator aIterator;
		aIterator=aVector.begin();
		while (aIterator!=aVector.end() &&
			   !pClass->IsFinished())
		{
			//Shoot the event
			pClass->SocketDone(*aIterator,
							   !pClass->m_bDefaultDisconnected);

			//Next
			++aIterator;
		}

		//Done
		return TRUE;
	}
	ERROR_HANDLER_STATIC_RETURN(CTCPPortScannerStealth_Class,"TimeoutProc",FALSE)
}

void CTCPPortScannerStealth::ResetPortsData()
{
	try
	{
		//Clear the map
		m_aPortsMap.clear();
	}
	ERROR_HANDLER("ResetPortsData")
}

void CTCPPortScannerStealth::StartThread(BOOL bDefaultDisconnected)
{
	try
	{
		//Set disconnection mode
		m_bDefaultDisconnected=bDefaultDisconnected;

		//Create the thread
		m_pThread=new CPeriodicThread(TimeoutProc);
		m_pThread->Start(500,(LPVOID)this);
	}
	ERROR_HANDLER("StartThread")
}

KOMODIA_NAMESPACE_END
