/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "UDPScanner.h"

#include "ErrorHandlerMacros.h"

#include "UDPSocket.h"

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

KOMODIA_NAMESPACE_START

// ----------------------------- CTCPScanner start ---------------------

#define CTCPScanner_Class "CTCPScanner"

CUDPScanner::CTCPScanner::CTCPScanner() : CTCPSocketAsync(TRUE),
										  m_pSocket(NULL)
{
	try
	{
		//Set our name
		SetName(CTCPScanner_Class);

		//Create the socket
		m_pSocket=new CUDPSocket(FALSE);
	}
	ERROR_HANDLER("CTCPScanner")
}

CUDPScanner::CTCPScanner::~CTCPScanner()
{
	try
	{
		delete m_pSocket;
	}
	ERROR_HANDLER("~CTCPScanner")
}

BOOL CUDPScanner::CTCPScanner::Create()
{
	try
	{
		//Delegate the call to the UDP socket
		return m_pSocket->Create();
	}
	ERROR_HANDLER_RETURN("Create",FALSE)
}

BOOL CUDPScanner::CTCPScanner::Connect(unsigned short usSourcePort,
									   IP aDestinationAddress,
									   unsigned short usDestinationPort,
									   BOOL bDisableAsync,
									   BOOL bForceErrorEvent)
{
	try
	{
		//Delegate the call
		return m_pSocket->Send(usSourcePort,
							   aDestinationAddress,
							   usDestinationPort,
							   NULL,
							   0)!=GetErrorCode();
	}
	ERROR_HANDLER_RETURN("Connect",FALSE)
}

// ----------------------------- CTCPScanner end ---------------------

// ----------------------------- CSniffingSocket start ---------------------

#define CSniffingSocket_Class "CSniffingSocket"

CUDPScanner::CSniffingSocket::CSniffingSocket(CUDPScanner* pScanner) : CICMPSocketAsync(TRUE),
																	   m_pFather(pScanner)
{
	try
	{
		//Set our name
		SetName(CSniffingSocket_Class);
	}
	ERROR_HANDLER("CSniffingSocket")
}

CUDPScanner::CSniffingSocket::~CSniffingSocket()
{
}

BOOL CUDPScanner::CSniffingSocket::OnSocketReceive(int iErrorCode)
{
	try
	{
		//First check if no error code
		if (iErrorCode)
			return TRUE;

		//Let parent proccess
		if (CICMPSocketAsync::OnSocketReceive(iErrorCode))
			//Check that this is the correct responce
			if (GetLastICMPHeader()->ucICMPType==ICMP_Unreachable && 
				GetLastIPHeader()->ulSourceAddress==m_pFather->GetTarget())
			{
				//Get the destination port
				UDPHeader aHeader;
				memcpy(&aHeader,
					   GetLastData(),
					   ICMP_DATA_SIZE);

				//Get the port
				unsigned short usPort;
				usPort=htons(aHeader.usDestinationPort);
				
				//Remove from list
				m_pFather->RemovePort(usPort);

				//Indicate status
				m_pFather->SocketDone(usPort,
									  FALSE);
				//Done
				return TRUE;
			}
			
		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("OnSocketReceive",FALSE)
}
// ----------------------------- CSniffingSocket end ---------------------

#define CUDPScanner_Class "CUDPScanner"

CUDPScanner::CUDPScanner(IP aBindInterface) : CTCPPortScannerStealth(aBindInterface),
											  m_pScanner(NULL),
											  m_pSniffer(NULL)
{
	try
	{
		//Set our name
		SetName(CUDPScanner_Class);
	}
	ERROR_HANDLER("CUDPScanner")
}

CUDPScanner::CUDPScanner(const std::string& rBindInterface) : CTCPPortScannerStealth(rBindInterface),
															  m_pScanner(NULL),
															  m_pSniffer(NULL)
{
	try
	{
		//Set our name
		SetName(CUDPScanner_Class);
	}
	ERROR_HANDLER("CUDPScanner")
}

CUDPScanner::~CUDPScanner()
{
	try
	{
		//Delete the scanner socket
		delete m_pScanner;

		//Delete the sniffer
		delete m_pSniffer;
	}
	ERROR_HANDLER("~CUDPScanner")
}

BOOL CUDPScanner::Scan(const std::string& rDestinationAddress)
{
	try
	{
		return Scan(CSpoofBase::StringToLong(rDestinationAddress));
	}
	ERROR_HANDLER_RETURN("Scan",FALSE)
}

BOOL CUDPScanner::Scan(IP aTarget)
{
	//Check max sockets are ok
	if (GetMaxSockets()<=0 ||
		!IsDone() ||
		!aTarget)
	{
		//Report it
		ReportError("Scan","Invalid parameters!");

		//Exit
		return FALSE;
	}

	try
	{
		//Reset the pair
		ResetPair();

		//Can we scan
		if (!CanScan())
		{
			//Error
			ReportError("Scan","No ports given!");

			//Error
			SetError(TRUE);

			//Exit
			return FALSE;
		}

		//Clear the ports map
		ResetPortsData();

		//Do we have the socket?
		if (m_pScanner)
			delete m_pScanner;

		//Allocate it
		m_pScanner=new CTCPScanner;

		//Try to create it
		if (!m_pScanner->Create())
		{
			//Report it
			ReportError("Scan","Failed to create socket!");

			//Exit
			return FALSE;
		}

		//Try to create the sniffer
		if (!CreateSniffer())
		{
			//Error
			ReportError("Scan","Failed to create sniffer!");

			//Error
			SetError(TRUE);

			//Exit
			return FALSE;
		}

		//Set the flags
		SetError(FALSE);
		SetDone(FALSE);

		//Save the number of max sockets
		ResetSocketCount();

		//Delete the port data
		DeletePortsMap();

		//Delete other data
		ResetData();

		//Set our target
		SetTarget(aTarget);

		//Create the thread
		StartThread(FALSE);

		//Start the loop
		for (int iCounter=0;
			 iCounter<GetMaxSockets();
			 ++iCounter)
		{
			BOOL bQuit;
			bQuit=FALSE;

			//Try to allocate sockets
			while (!bQuit)
			{
				//Another socket
				AdjustNumberOfSockets(1);

				//Create it
				if (!(bQuit=NewSocket()))
					ReportError("Scan","Failed to create socket!");
			}
		}

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("Scan",FALSE)
}

CTCPSocketAsync* CUDPScanner::AllocateSocket(unsigned short usPort)const
{
	return m_pScanner;
}

void CUDPScanner::DestroySocket(CTCPSocketAsync* pSocket)const
{
}

BOOL CUDPScanner::CreateSniffer()
{
	try
	{
		//Try to create the sniffer socket
		if (m_pSniffer)
		{
			delete m_pSniffer;
			m_pSniffer=NULL;
		}

		//Create it	
		CSniffingSocket* pSocket;
		pSocket=new CSniffingSocket(this);

		//Protect it
		std::auto_ptr<CSniffingSocket> pProtection(pSocket);

		//Continue to create it
		if (!pSocket->Create())
		{
			//Report it
			ReportError("CreateSniffer","Failed to create sniffer");

			//Exit
			return FALSE;
		}
		
		//Try to bind it
		if (!pSocket->Bind(GetBindInterface(),0))
		{
			//Report it
			ReportError("CreateSniffer","Failed to bind sniffer");

			//Exit
			return FALSE;
		}

		//Release protection and save socket
		m_pSniffer=pProtection.release();

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("CreateSniffer",FALSE)
}

void CUDPScanner::TCPScanDoneLib(BOOL bError)
{
	try
	{
		//Delete the socket
		CSniffingSocket* pSocket;
		pSocket=m_pSniffer;	
		m_pSniffer=NULL;

		//From it's thread
		pSocket->DeleteSocketFromThread();

		//Delegate call
		CTCPPortScannerStealth::TCPScanDoneLib(bError);
	}
	ERROR_HANDLER("TCPScanDoneLib")
}

KOMODIA_NAMESPACE_END
