/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "SpoofSocket.h"

#include "ErrorHandlerMacros.h"
#include "Array_ptr.h"

#include <memory>

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

KOMODIA_NAMESPACE_START

//----------------------- CIPOptions start -----------------------

#define CIPOptions_LOGNAME "CIPOptions"

CIPOptions::CIPOptions() : CSpoofBase(),
						   m_pBuffer(NULL),
						   m_bAutoPAD(TRUE)
{
	try
	{
		//Set our name
		SetName(CIPOptions_LOGNAME);

		//Initialize our buffer
		m_pBuffer=new char[IPOption_SIZE];

		//Set our buffer to nothing
		Reset();
	}
	ERROR_HANDLER("CIPOptions")
}

CIPOptions::CIPOptions(const CIPOptions& rIP) : CSpoofBase(rIP),
											    m_pBuffer(NULL),
											    m_bAutoPAD(rIP.m_bAutoPAD)
{
	try
	{
		//Set our name
		SetName(CIPOptions_LOGNAME);

		//Initialize our buffer
		m_pBuffer=new char[IPOption_SIZE];

		//Set our buffer to nothing
		Reset();

		//Copy the data
		memcpy(m_pBuffer,
			   rIP.m_pBuffer,
			   IPOption_SIZE);

		//Take the length
		m_iBufferLength=rIP.m_iBufferLength;
	}
	ERROR_HANDLER("CIPOptions")
}

CIPOptions::~CIPOptions()
{
	try
	{
		delete [] m_pBuffer;
	}
	ERROR_HANDLER("~CIPOptions")
}

void CIPOptions::AddOption_Nothing()
{
	try
	{
		//Add option do nothing
		OptionType aOptionType;

		//Get the option
		aOptionType=GetOption(IPOption_DONT_COPY,
							  IPOption_CONTROL,
							  IPOption_NO_OPERATION);

		//Add it to buffer
		AddToBuffer((char*)&aOptionType,sizeof(aOptionType));
	}
	ERROR_HANDLER("AddOption_Nothing")
}

CIPOptions::OptionType CIPOptions::GetOption(unsigned char ucCopyFlag, 
											 unsigned char ucClassFlag, 
											 unsigned char ucTypeFlag)
{
	//Return a single option type
	return ucCopyFlag | 
		   ucClassFlag | 
		   ucTypeFlag;
}

void CIPOptions::AddToBuffer(const char* pBuffer,
							 int iBufLength)
{
	if (m_iBufferLength<IPOption_SIZE)
	{
		//Add our option to the buffer
		memcpy(m_pBuffer+m_iBufferLength,
			   pBuffer,
			   iBufLength);

		//Adjust the size
		m_iBufferLength+=iBufLength;
	}
}

const char* CIPOptions::GetBuffer()const
{
	return m_pBuffer;
}

int CIPOptions::GetBufferLength()const
{
	try
	{
		//Check if auto pad or not
		if (m_bAutoPAD)
			if (m_iBufferLength/IPOption_WRAPSIZE==(m_iBufferLength/IPOption_WRAPSIZE)*IPOption_WRAPSIZE && m_iBufferLength>=IPOption_WRAPSIZE)
				return m_iBufferLength;
			else
				return int((float)m_iBufferLength/IPOption_WRAPSIZE+1)*IPOption_WRAPSIZE;
		else
			return m_iBufferLength;
	}
	ERROR_HANDLER_RETURN("GetBufferLength",0)
}

void CIPOptions::AddOption_ENDLIST()
{
	try
	{
		//End the list of options
		OptionType aOptionType;

		//Get the option
		aOptionType=GetOption(IPOption_DONT_COPY,
							  IPOption_CONTROL,
							  IPOption_END_OPTION);

		//Add it to buffer
		AddToBuffer((char*)&aOptionType,sizeof(aOptionType));
	}
	ERROR_HANDLER("AddOption_ENDLIST")
}

void CIPOptions::SetAutoPad(BOOL bAutoPAD)
{
	m_bAutoPAD=bAutoPAD;
}

CIPOptions* CSpoofSocket::GetOptions()const
{
	return m_pIPOptions;
}

void CIPOptions::Reset()
{
	try
	{
		//Set all to zeros
		memset(m_pBuffer,
			   0,
			   IPOption_SIZE);

		//Our buffer length
		m_iBufferLength=0;
	}
	ERROR_HANDLER("Reset")
}

void CIPOptions::AddOption_Security(unsigned short usType)
{
	try
	{
		//Add option security
		OptionType aOptionType;

		//Get the option
		aOptionType=GetOption(IPOption_COPY,
							  IPOption_CONTROL,
							  IPOption_SECURITY);

		//Add it to buffer
		AddToBuffer((char*)&aOptionType,
					sizeof(aOptionType));

		//Add length
		aOptionType=IPOption_SECURITY_LENGTH;
		AddToBuffer((char*)&aOptionType,
					sizeof(aOptionType));

		//Add options
		AddToBuffer((char*)&usType,
					sizeof(usType));

		//Add zeros
		unsigned short usZeros;
		usZeros=0;
		unsigned char ucZeros;
		ucZeros=0;

		//A hacker would enumarate these values, according to the RFC
		//Compartments
		AddToBuffer((char*)&usZeros,
					sizeof(usZeros));

		//Handling restrictions
		AddToBuffer((char*)&usZeros,
					sizeof(usZeros));

		//Transmition control code (TCC)
		AddToBuffer((char*)&usZeros,
					sizeof(usZeros));
		AddToBuffer((char*)&ucZeros,
					sizeof(ucZeros));

		//Done
	}
	ERROR_HANDLER("AddOption_Security")
}

void CIPOptions::AddOption_Stream(unsigned short usStreamID)
{
	try
	{
		//Add option security
		OptionType aOptionType;

		//Get the option
		aOptionType=GetOption(IPOption_COPY,
							  IPOption_CONTROL,
							  IPOption_STREAM);

		//Add it to buffer
		AddToBuffer((char*)&aOptionType,
					sizeof(aOptionType));

		//Add length
		aOptionType=IPOption_STREAM_LENGTH;
		AddToBuffer((char*)&aOptionType,
					sizeof(aOptionType));
		
		//Convert endian format
		unsigned short usnStreamID;
		usnStreamID=htons(usStreamID);
		
		//Add options
		AddToBuffer((char*)&usnStreamID,
					sizeof(usnStreamID));
	}
	ERROR_HANDLER("AddOption_Stream")
}

void CIPOptions::AddOption_StrictRoute(Routing tRoute)
{
	try
	{
		AddOption_Route(IPOption_STRICT_ROUTING,tRoute);
	}
	ERROR_HANDLER("AddOption_StrictRoute")
}

void CIPOptions::AddOption_RecordRoute(int iMaxRoutes)
{
	try
	{
		//Option for strict routine
		//Add option strict route
		OptionType aOptionType;

		//Get the option
		aOptionType=GetOption(IPOption_DONT_COPY,
							  IPOption_CONTROL,
							  IPOption_RECORD_ROUTE);

		//Add it to buffer
		AddToBuffer((char*)&aOptionType,
					sizeof(aOptionType));

		//Add the length
		aOptionType=iMaxRoutes*4+IPOption_STRICT_ROUTING_LENGTH;
		AddToBuffer((char*)&aOptionType,
					sizeof(aOptionType));
		
		//Add the pointer
		aOptionType=IPOption_STRICT_ROUTING_POINTER;
		AddToBuffer((char*)&aOptionType,
					sizeof(aOptionType));
		
		char cNothing[IPOption_SIZE]="";
		AddToBuffer(cNothing,
					iMaxRoutes*4);
	}
	ERROR_HANDLER("AddOption_RecordRoute")
}

void CIPOptions::AddOption_Route(OptionType aRouteType,
								 Routing aRoute)
{
	try
	{
		//Option for strict routine
		//Add option strict route
		OptionType aOptionType;

		//Get the option
		aOptionType=GetOption(IPOption_COPY,
							  IPOption_CONTROL,
							  aRouteType);

		//Add it to buffer
		AddToBuffer((char*)&aOptionType,
					sizeof(aOptionType));

		//Add the length
		aOptionType=aRoute.iRoutes*4+IPOption_STRICT_ROUTING_LENGTH;
		AddToBuffer((char*)&aOptionType,
					sizeof(aOptionType));
		
		//Add the pointer
		aOptionType=IPOption_STRICT_ROUTING_POINTER;
		AddToBuffer((char*)&aOptionType,
					sizeof(aOptionType));
		
		//Add the routing table
		AddToBuffer((char*)aRoute.ulRoutes,
					aRoute.iRoutes*4);
	}
	ERROR_HANDLER("AddOption_Route")
}

void CIPOptions::AddOption_LooseRoute(Routing tRoute)
{
	try
	{
		AddOption_Route(IPOption_LOOSE_ROUTING,tRoute);
	}
	ERROR_HANDLER("AddOption_LooseRoute")
}

void CIPOptions::AddOption_Timestamp(OptionType tFlags, int iMaxStamps)
{
	try
	{
		//Add option for timestamp
		OptionType aOptionType;

		//Get the option
		aOptionType=GetOption(IPOption_DONT_COPY,
							  IPOption_DEBUGGING,
							  IPOption_TIMESTAMP);

		//Add it to buffer
		AddToBuffer((char*)&aOptionType,
					sizeof(aOptionType));

		//Add the length
		aOptionType=iMaxStamps*IPOption_TIMESTAMP_SIZE+IPOption_TIMESTAMP_LENGTH-1;
		AddToBuffer((char*)&aOptionType,
					sizeof(aOptionType));
		
		//Add the pointer
		aOptionType=IPOption_TIMESTAMP_LENGTH;
		AddToBuffer((char*)&aOptionType,
					sizeof(aOptionType));

		//Add the flags
		AddToBuffer((char*)&tFlags,
					sizeof(tFlags));

		//Add the empty buffer
		char cNothing[IPOption_SIZE]="";
		AddToBuffer(cNothing,
					iMaxStamps*IPOption_TIMESTAMP_SIZE);
	}
	ERROR_HANDLER("AddOption_Timestamp")
}

//----------------------- CIPOptions end -----------------------

#define CSpoofSocket_LOGNAME "CSpoofSocket"

CSpoofSocket::CSpoofSocket(BOOL bRawSocket) : CSpoofBase(),
											  m_bRaw(bRawSocket)
{
	try
	{
		//Set our name
		SetName(CSpoofSocket_LOGNAME);

		//Initialize the data
		InitializeIP();
	}
	ERROR_HANDLER("CSpoofSocket")
}

CSpoofSocket::CSpoofSocket(SOCKET aSocket) : CSpoofBase()
{
	//Check it's a valid socket
	if (aSocket!=INVALID_SOCKET)
		ReportError("CSpoofSocket","Received invalid socket!");
	else
		try
		{
			//Set our name
			SetName(CSpoofSocket_LOGNAME);

			//Initialize the data
			InitializeIP();

			//Save the handle
			AssignSocket(aSocket);
		}
		ERROR_HANDLER("CSpoofSocket")
}

CSpoofSocket::~CSpoofSocket()
{
	try
	{
		//Delete options
		SetOptions(FALSE);

		//Close the socket
		Close();
	}
	ERROR_HANDLER("~CSpoofSocket")
}

BOOL CSpoofSocket::Create(int iProtocol)
{
	//Close the socket if open
	if (ValidSocket())
		Close();

	try
	{
		//Are we overlapped
		if (!m_bOverlapped)
		{
			//Here we create the raw socket
			if (m_bRaw || iProtocol==IPPROTO_ICMP)
				m_aSpoofSocket=socket(AF_INET,
									  SOCK_RAW,
									  iProtocol);
			else
				if (iProtocol==IPPROTO_TCP)
					m_aSpoofSocket=socket(AF_INET,
										  SOCK_STREAM,
										  iProtocol);
				else if (iProtocol==IPPROTO_UDP)
					m_aSpoofSocket=socket(AF_INET,
										  SOCK_DGRAM,
										  iProtocol);
		}
		else
		{
			//Here we create the raw socket
			if (m_bRaw || iProtocol==IPPROTO_ICMP)
				m_aSpoofSocket=WSASocket(AF_INET,
										 SOCK_RAW,
										 iProtocol,
										 NULL,
										 NULL,
										 WSA_FLAG_OVERLAPPED);
			else
				if (iProtocol==IPPROTO_TCP)
					m_aSpoofSocket=WSASocket(AF_INET,
											 SOCK_STREAM,
											 iProtocol,
											 NULL,
											 NULL,
											 WSA_FLAG_OVERLAPPED);
				else if (iProtocol==IPPROTO_UDP)
					m_aSpoofSocket=WSASocket(AF_INET,
											 SOCK_DGRAM,
											 iProtocol,
											 NULL,
											 NULL,
											 WSA_FLAG_OVERLAPPED);
		}

		//Check for socket validity
		if (m_aSpoofSocket==INVALID_SOCKET)
		{
			//Error
			SetLastError("Create");

			//Done
			return FALSE;
		}

		if (m_bRaw)
		{
			//Set that the application will send the IP header
			unsigned int iTrue=1;

			if(setsockopt(m_aSpoofSocket,
						  IPPROTO_IP,
						  IP_HDRINCL,
						  (char*)&iTrue,
						  sizeof(iTrue))==GetErrorCode())
			{
				//Check for options error
				SetLastError("Create");

				//Exit
				return FALSE;
			}
		}

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("Create",FALSE)
}

int CSpoofSocket::SendRawBuffer(IP aDestinationAddress,
								const char* pBuffer,
								unsigned long ulBufferLength,
								unsigned short usDestinationPort)
{
	try
	{
		//Quit if not ok
		//Make things the fastest
		if (m_aSpoofSocket==INVALID_SOCKET ||
			!m_bRaw)
			return FALSE;

		//Define the target address
		sockaddr_in m_TargetAddress;

		m_TargetAddress.sin_family=AF_INET;
		m_TargetAddress.sin_addr.s_addr=aDestinationAddress;
		m_TargetAddress.sin_port=htons(usDestinationPort);
		memset(&m_TargetAddress.sin_zero,0,sizeof(m_TargetAddress.sin_zero));

		//And send the data
		//Send the data
		int iResult;
		iResult=sendto(GetHandle(),
					   pBuffer,
					   ulBufferLength,
					   0,
					   (sockaddr*)&m_TargetAddress,
					   sizeof(m_TargetAddress));

		if (iResult==GetErrorCode())
			//Set the error
			SetLastError("SendRawBuffer");

		//Done
		return iResult;
	}
	ERROR_HANDLER_RETURN("SendRawBuffer",GetErrorCode())
}

int CSpoofSocket::Send(IP aDestinationAddress,
					   const char* pBuffer,
					   unsigned long ulBufferLength,
					   unsigned short usDestinationPort)
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return GetErrorCode();

		//Is it a valid size
		if (ulBufferLength &&
			!pBuffer)
			return GetErrorCode();

		//Define the target address
		sockaddr_in aTargetAddress;
		memset(&aTargetAddress,
			   0,
			   sizeof(aTargetAddress));

		aTargetAddress.sin_family=AF_INET;
		aTargetAddress.sin_addr.s_addr=aDestinationAddress;
		aTargetAddress.sin_port=htons(usDestinationPort);

		//packet send status ?
		int iResult;

		//Only if allowing raw headers !!
		if (m_bRaw)
		{
			//Header length
			unsigned char ucHeaderLength;
			ucHeaderLength=IpHeaderLength;

			//Do we have options?
			if (m_bOptions)
				ucHeaderLength+=m_pIPOptions->GetBufferLength();
			
			//First construct the packet
			LPIpHeader lpHead=ConstructIPHeader(m_ucProtocol,
												IpFragFlag_DONT_FRAG,
												m_ucTTL,
												(unsigned short)GetCurrentProcessId(),
												ucHeaderLength);

			//Protect the header
			std::auto_ptr<IpHeader> pProtection(lpHead);

			//Set the address
			SetIPHeaderAddress(lpHead,
							   m_ulSourceAddress,
							   aDestinationAddress);

			//Now add some more options
			int iTotalLength;
			iTotalLength=ucHeaderLength+ulBufferLength;

			//Set the header
			lpHead->usTotalLength=htons(iTotalLength);

			//Need to construct a new packet
			char* pNewBuffer;
			pNewBuffer=new char[iTotalLength];

			//Protect the buffer
			CArray_ptr<char> pBufferProtection(pNewBuffer);

			//Copy two buffers
			memcpy(pNewBuffer,
				   lpHead,
				   IpHeaderLength);

			//Do we need to copy options ?
			if (m_bOptions)
				memcpy(pNewBuffer+IpHeaderLength,
					   m_pIPOptions->GetBuffer(),
					   m_pIPOptions->GetBufferLength());

			//Only if not null
			if (pBuffer)
				memcpy(pNewBuffer+ucHeaderLength,
					   pBuffer,
					   ulBufferLength);
			
			//Calculate the checksum
			lpHead->usChecksum=CalculateChecksum((unsigned short*)pNewBuffer,
												 ucHeaderLength);

			//Alert everyone this is the final header
			FinalIPHeader(lpHead);

			//Recopy the ip
			memcpy(pNewBuffer,
				   lpHead,
				   IpHeaderLength);
			
			//Send the data
			iResult=sendto(GetHandle(),
						   (const char*)pNewBuffer,
						   iTotalLength,
						   0,
						   (sockaddr*)&aTargetAddress,
						   sizeof(aTargetAddress));

			//Is all OK
			if (iResult==GetErrorCode() &&
				GetSystemLastError()!=WSAEWOULDBLOCK)
				//Set the error
				SetLastError("Send - Raw");
		}
		else
		{
			//Set to no error
			iResult=!GetErrorCode();

			//Insert options
			/*if (m_bOptions)
				iResult=setsockopt(GetHandle(),
								   IPPROTO_IP,
								   IP_OPTIONS,
								   m_pIPOptions->GetBuffer(),
								   m_pIPOptions->GetBufferLength());
			else
				//No options
				iResult=setsockopt(GetHandle(),
								   IPPROTO_IP,
								   IP_OPTIONS,
								   NULL,
								   0);*/

			//Check if we had an error
			if (iResult!=GetErrorCode())
				//Use regular send !!!
				iResult=sendto(GetHandle(),
							   (const char*)pBuffer,
							   ulBufferLength,
							   0,
							   (sockaddr*)&aTargetAddress,
							   sizeof(aTargetAddress));
			
			//Is all OK?
			if (iResult==GetErrorCode() &&
				GetSystemLastError()!=WSAEWOULDBLOCK)
				//Set the error
				SetLastError("Send");
		}
		
		//Done
		return iResult;
	}
	ERROR_HANDLER_RETURN("Send",GetErrorCode())
}

int CSpoofSocket::Send(const std::string& rDestinationAddress,
					   const char* pBuffer,
					   unsigned long ulBufferLength,
					   unsigned short usDestinationPort)
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return GetErrorCode();

		return Send(inet_addr(rDestinationAddress.c_str()),
					pBuffer,
					ulBufferLength,
					usDestinationPort);
	}
	ERROR_HANDLER_RETURN("Send",GetErrorCode())
}

LPIpHeader CSpoofSocket::ConstructIPHeader(unsigned char  ucProtocol,
										   unsigned short usFragmentationFlags,
										   unsigned char  ucTTL,
										   unsigned short usIdentification,
										   unsigned char  ucHeaderLength)const
{
	return ConstructStaticIPHeader(ucProtocol,
								   usFragmentationFlags,
								   ucTTL,
								   usIdentification,
								   ucHeaderLength);				   
}

void CSpoofSocket::SetIPHeaderAddress(LPIpHeader lpHead,
									  IP aSourceAddress,
									  IP aDestinationAddress)const
{
	try
	{
		//We need to place the header
		//If source is NULL then we need to use default source
		if (!aSourceAddress || !aDestinationAddress)
		{
			//Report it
			ReportError("SetIPHeaderAddress","Recieved empty source or destination address!");

			//Exit
			return;
		}

		//Place source address
		lpHead->ulSourceAddress=aSourceAddress;

		//Place destination address
		lpHead->ulDestinationAddress=aDestinationAddress;

		//Done
	}
	ERROR_HANDLER("SetIPHeaderAddress")
}

BOOL CSpoofSocket::ValidSocket() const
{
	return m_aSpoofSocket!=INVALID_SOCKET;
}

unsigned short CSpoofSocket::CalculateChecksum(const unsigned short *pBuffer, 
											   int iSize)
{
	try
	{
		unsigned long usChksum=0;

		//Calculate the checksum
		while (iSize>1)
		{
			usChksum+=*pBuffer++;
			iSize-=sizeof(unsigned short);
		}

		//If we have one char left
		if (iSize)
			usChksum+=*(unsigned char*)pBuffer;

		//Complete the calculations
		usChksum=(usChksum >> 16) + (usChksum & 0xffff);
		usChksum+=(usChksum >> 16);

		//Return the value (inversed)
		return (unsigned short)(~usChksum);
	}
	ERROR_HANDLER_STATIC_RETURN(CSpoofSocket_LOGNAME,"CalculateChecksum",0)
}

BOOL CSpoofSocket::Bind(IP aSourceAddress,
						unsigned short usPort)
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return FALSE;

		//Create the local address
		sockaddr_in soSrc;

		//Set to 0
		memset(&soSrc,0,sizeof(soSrc));
		soSrc.sin_family=AF_INET;

		//Populate the connection data
		if (aSourceAddress)
			soSrc.sin_addr.s_addr=aSourceAddress;
		else
			soSrc.sin_addr.s_addr=ADDR_ANY ;

		soSrc.sin_port=htons(usPort);

		//Now we need to bind it
		if (bind(GetHandle(),
				 (sockaddr*)&soSrc,
				 sizeof(soSrc)))
		{
			//Error
			SetLastError("Bind");

			//Exit
			return FALSE;
		}
		else
			//Save the address
			m_aConnectedTo=soSrc;

		//If already has a source address then don't change it
		if (!m_ulSourceAddress)
			//Save it as the source address (spoofing options only)
			m_ulSourceAddress=aSourceAddress;

		return TRUE;
	}
	ERROR_HANDLER_RETURN("Bind",FALSE)
}

BOOL CSpoofSocket::Bind(const std::string& rSourceAddress,
						unsigned short usPort)
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return FALSE;

		if (rSourceAddress.empty())
			return Bind((IP)0,
						usPort);
		else
			return Bind(inet_addr(rSourceAddress.c_str()),
						usPort);
	}
	ERROR_HANDLER_RETURN("Bind",FALSE)
}

SOCKET CSpoofSocket::GetHandle() const
{
	return m_aSpoofSocket;
}

BOOL CSpoofSocket::CheckSocketValid() const
{
	try
	{
		//Check if socket is invalid
		if (!ValidSocket())
		{
			//Report it
			ReportError("CheckSocketValid","Operation made on non existant socket!");

			//Exit
			return FALSE;
		}	

		//OK
		return TRUE;
	}
	ERROR_HANDLER_RETURN("CheckSocketValid",FALSE)
}

BOOL CSpoofSocket::Close()
{
	try
	{
		//Close the socket
		//Quit if not ok
		if (!ValidSocket())
			return FALSE;

		//Close it
		if (closesocket(GetHandle())==GetErrorCode())
		{
			//Error in closing ?
			SetLastError("Close");

			//Exit
			return FALSE;
		}

		//Set the socket to invalid
		m_aSpoofSocket=INVALID_SOCKET;

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("Close",FALSE)
}


void CSpoofSocket::SetProtocol(unsigned char ucProtocol)
{
	m_ucProtocol=ucProtocol;
}

void CSpoofSocket::SetSourceAddress(IP aSourceAddress)
{
	m_ulSourceAddress=aSourceAddress;
}

void CSpoofSocket::SetSourceAddress(const std::string& rSourceAddress)
{
	try
	{
		//Set the source address, in case we want to spoof it
		if (rSourceAddress.empty())
			m_ulSourceAddress=0;
		else
			m_ulSourceAddress=inet_addr(rSourceAddress.c_str());
	}
	ERROR_HANDLER("SetSourceAddress")
}

unsigned short CSpoofSocket::CalculatePseudoChecksum(const char* pBuffer, 
													 int iBufferLength,
													 IP aDestinationAddress,
													 int iPacketLength)const
{
	try
	{
		//Calculate the checksum
		LPPseudoHeader lpPseudo;
		lpPseudo=new PseudoHeader;

		//Protect it
		std::auto_ptr<PseudoHeader> pProtection(lpPseudo);

		//Set the values
		lpPseudo->ulDestinationAddress=aDestinationAddress;
		lpPseudo->ulSourceAddress=m_ulSourceAddress;
		lpPseudo->ucZeros=0;
		lpPseudo->ucPTCL=m_ucProtocol;
		lpPseudo->usLength=htons(iPacketLength);

		//Calculate checksum of all
		int iTotalLength;
		iTotalLength=PseudoHeaderLength+iBufferLength;

		//Allocate the buffer
		char* pNewBuffer;
		pNewBuffer=new char[iTotalLength];

		//Protect the new buffer
		CArray_ptr<char> pBufferProtection(pNewBuffer);

		//Copy pseudo
		memcpy(pNewBuffer,lpPseudo,PseudoHeaderLength);

		//Copy header
		memcpy(pNewBuffer+PseudoHeaderLength,
			   pBuffer,
			   iBufferLength);

		//Calculate the checksum
		unsigned short usChecksum;
		usChecksum=CalculateChecksum((unsigned short*)pNewBuffer,iTotalLength);

		//Return checksum
		return usChecksum;
	}
	ERROR_HANDLER_RETURN("CalculatePseudoChecksum",0)
}

void CSpoofSocket::SetTTL(unsigned char ucTTL)
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return;

		if (m_bRaw)
			//Set the ttl
			m_ucTTL=ucTTL;
		else if(setsockopt(GetHandle(),
						   IPPROTO_IP,
						   IP_TTL,
						   (const char*)&ucTTL,
						   sizeof(ucTTL)))
			SetLastError("SetTTL");
	}
	ERROR_HANDLER("SetTTL")
}

void CSpoofSocket::SetOptions(BOOL bOptions)
{
	try
	{
		//Do we want options, normaly not
		m_bOptions=bOptions;

		//Do we have older options?
		if (m_pIPOptions)
		{
			delete m_pIPOptions;
			m_pIPOptions=NULL;
		}

		//Do we need to reallocate the options
		if (bOptions)
			m_pIPOptions=new CIPOptions;
	}
	ERROR_HANDLER("SetOptions")
}

BOOL CSpoofSocket::IsRaw() const
{
	return m_bRaw;
}

void CSpoofSocket::InitializeIP()
{
	try
	{
		//Invalid the socket
		m_aSpoofSocket=INVALID_SOCKET;

		//More invalids
		m_ulSourceAddress=0;

		//Some defaults
		m_ucTTL=IP_DEF_TTL;

		//Set our options
		m_pIPOptions=NULL;

		//Not sniffing
		m_bSniffing=FALSE;

		//We are not overlapped
		m_bOverlapped=FALSE;

		//Not connected
		memset(&m_aConnectedTo,
			   0,
			   sizeof(m_aConnectedTo));

		//Set options to false
		SetOptions(FALSE);
	}
	ERROR_HANDLER("InitializeIP")
}

void CSpoofSocket::AssignSocket(SOCKET aNewSocket,unsigned char ucProtocol)
{
	try
	{
		//Sets the protocol
		m_ucProtocol=ucProtocol;

		//Binds to a socket
		m_aSpoofSocket=aNewSocket;
	}
	ERROR_HANDLER("AssignSocket")
}

int CSpoofSocket::ReceiveFrom(char* pBuffer,
							  unsigned long ulBufferLength,
							  IP& rIP,
							  unsigned short& rSourcePort)
{
	try
	{
		if (!ValidSocket() ||
			!pBuffer ||
			!ulBufferLength)
			return GetErrorCode();

		//Receive data
		int iResult;

		//Receive
		if (m_ucProtocol!=IPPROTO_TCP && 
			!m_bSniffing)
		{
			//Get the remote address
			sockaddr saConnected;

			int iTmp;
			iTmp=sizeof(saConnected);

			//Accept it
			iResult=recvfrom(GetHandle(),
							 pBuffer,
							 ulBufferLength,
							 NULL,
							 &saConnected,
							 &iTmp);

			//If OK set it
			if (iResult!=GetErrorCode())
			{
				//Address
				rIP=((sockaddr_in*)&saConnected)->sin_addr.S_un.S_addr;

				//Port
				rSourcePort=htons(((sockaddr_in*)&saConnected)->sin_port);

				//Done
				return iResult;
			}
			else
			{
				//Error
				SetLastError("Receive");

				//Reset the data
				rIP=0;
				rSourcePort=0;

				//Done
				return iResult;
			}
		}
		else
		{
			//Report it
			ReportError("Receive","Can't run on TCP socket!");

			//Exit
			return GetErrorCode();
		}
	}
	ERROR_HANDLER_RETURN("Receive",GetErrorCode())
}

int CSpoofSocket::Receive(char *pBuffer, 
						  unsigned long ulBufferLength)
{
	try
	{
		if (!ValidSocket() ||
			!pBuffer ||
			!ulBufferLength)
			return GetErrorCode();

		//Receive data
		int iResult;

		//Receive
		if (m_ucProtocol!=IPPROTO_TCP && 
			!m_bSniffing)
		{
			//Get the remote address
			sockaddr saConnected;

			int iTmp;
			iTmp=sizeof(saConnected);

			//Accept it
			iResult=recvfrom(GetHandle(),
							 pBuffer,
							 ulBufferLength,
							 NULL,
							 &saConnected,
							 &iTmp);

			//If OK set it
			if (iResult!=GetErrorCode())
				//Take the connected to data
				memcpy(&m_aConnectedTo,
					   &saConnected,
					   sizeof(saConnected));
		}
		else
			//Do a regular receive
			iResult=recv(GetHandle(),
						 pBuffer,
						 ulBufferLength,
						 NULL);

		//Check if error
		if (iResult==GetErrorCode() &&
			GetSystemLastError()!=WSAEWOULDBLOCK)
			//Error
			SetLastError("Receive");

		//Number of bytes received
		return iResult;
	}
	ERROR_HANDLER_RETURN("Receive",GetErrorCode())
}

BOOL CSpoofSocket::ValidAddress(const std::string& rAddress)
{
	try
	{
		return inet_addr(rAddress.c_str())!=INADDR_NONE;
	}
	ERROR_HANDLER_STATIC_RETURN(CSpoofSocket_LOGNAME,"ValidAddress",FALSE)
}

sockaddr_in CSpoofSocket::InternalResolveDNS(const std::string& rAddress)
{
	//Convert it to the address
	sockaddr_in aAddr;
	memset(&aAddr,0,sizeof(aAddr));

	try
	{
		//Resolve the DNS
		hostent* pData;
		pData=gethostbyname(rAddress.c_str());

		//Check if this address exists
		if (pData)
			//Copy the data
			memcpy(&aAddr.sin_addr,
				   pData->h_addr,
				   pData->h_length);

		return aAddr;
	}
	ERROR_HANDLER_STATIC_RETURN(CSpoofSocket_LOGNAME,"InternalResolveDNS",aAddr)
}

long CSpoofSocket::ResolveDNS(const std::string& rAddress)
{
	try
	{
		//Resolve the DNS
		sockaddr_in aAddr;
		aAddr=InternalResolveDNS(rAddress.c_str());

		//Check if valid
		if (aAddr.sin_addr.S_un.S_addr==0)
			//Error
			return 0;
		else
			return aAddr.sin_addr.S_un.S_addr;
	}
	ERROR_HANDLER_STATIC_RETURN(CSpoofSocket_LOGNAME,"ResolveDNS",0)
}

BOOL CSpoofSocket::Sniff(BOOL bSniff)
{
	//Start sniffing
	if (!ValidSocket())
		return FALSE;

	try
	{
#ifdef WIN32
		unsigned long ulBytes;
		if (WSAIoctl(GetHandle(),SIO_RCVALL,&bSniff,sizeof(bSniff),NULL,0,&ulBytes,NULL,NULL))
		{
			//Error
			SetLastError("Sniff");

			//Exit
			return FALSE;
		}

		//Set sniffer status
		m_bSniffing=bSniff;

		//Done
		return TRUE;
#else
		return FALSE;
#endif
	}
	ERROR_HANDLER_RETURN("Sniff",FALSE)
}

long CSpoofSocket::GetPeerAddress() const
{
	//Get the address we are connected to
	return m_aConnectedTo.sin_addr.S_un.S_addr;
}

BOOL CSpoofSocket::Shutdown(SocketShutdown eHow)
{
	if (!CheckSocketValid())
		return FALSE;

	try
	{
		int iHow;

		//Convert the how to a real flag
		if (eHow==ssReceive)
			iHow=SD_RECEIVE;
		else if (eHow==ssSend)
			iHow=SD_SEND;
		else
			iHow=SD_BOTH;

		//Do it
		if (shutdown(GetHandle(),iHow))
		{
			//Report it
			SetLastError("Shutdown");

			//Exit
			return FALSE;
		}

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("Shutdown",FALSE)
}

unsigned short CSpoofSocket::GetPeerPort() const
{
	return htons(m_aConnectedTo.sin_port);
}

void CSpoofSocket::FinalIPHeader(LPIpHeader lpHead)const
{
	//We don't do anything
}

LPIpHeader CSpoofSocket::ConstructStaticIPHeader(unsigned char  ucProtocol,
												 unsigned short usFragmentationFlags,
												 unsigned char  ucTTL,
												 unsigned short usIdentification,
												 unsigned char  ucHeaderLength)
{
	try
	{
		//Need to construct the IP header
		LPIpHeader lpHead=new _IpHeader;

		//Header length (in 32 bits)
		lpHead->ucHeaderLength_Version=(ucHeaderLength >> 2) |
									   (IpVersion << 4);

		//Protocol
		lpHead->ucProtocol=ucProtocol;

		//Fragmentation flags
		lpHead->usFragmentationFlags=htons(usFragmentationFlags);

		//Time to live
		lpHead->ucTTL=ucTTL;

		//Checksum - set to 0
		lpHead->usChecksum=0;

		//Identification
		lpHead->usIdentification=htons(usIdentification);

		//Precedence
		lpHead->ucTypeOfService=IpService_ROUTINE;

		//Return IP to user
		return lpHead;
	}
	ERROR_HANDLER_STATIC_RETURN(CSpoofSocket_LOGNAME,"ConstructIPHeader",NULL)
}

BOOL CSpoofSocket::CanRead()const
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return FALSE;

		//Create our structure
		fd_set aDescriptor;
		FD_ZERO(&aDescriptor);

		//Add our socket
		FD_SET(GetHandle(),&aDescriptor);

		//And create the timeval
		timeval aTime;
		aTime.tv_sec=0;
		aTime.tv_usec=0;

		//And run the select
		int iRetVal;
		iRetVal=select(NULL,&aDescriptor,NULL,NULL,&aTime);

		//Check if we had an error
		if (iRetVal==GetErrorCode())
		{
			//Report it
			SetLastError("CanRead");

			//Exit
			return FALSE;
		}
		else
			//Check is our socket set
			return FD_ISSET(GetHandle(),&aDescriptor);
	}
	ERROR_HANDLER_RETURN("CanRead",FALSE)
}

BOOL CSpoofSocket::CanWrite()const
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return FALSE;

		//Create our structure
		fd_set aDescriptor;
		FD_ZERO(&aDescriptor);

		//Add our socket
		FD_SET(GetHandle(),&aDescriptor);

		//And create the timeval
		timeval aTime;
		aTime.tv_sec=0;
		aTime.tv_usec=0;

		//And run the select
		int iRetVal;
		iRetVal=select(NULL,NULL,&aDescriptor,NULL,&aTime);

		//Check if we had an error
		if (iRetVal==GetErrorCode())
		{
			//Report it
			SetLastError("CanWrite");

			//Exit
			return FALSE;
		}
		else
			//Check is our socket set
			return FD_ISSET(GetHandle(),&aDescriptor);
	}
	ERROR_HANDLER_RETURN("CanWrite",FALSE)
}

BOOL CSpoofSocket::IsCreated()const
{
	return ValidSocket();
}

BOOL CSpoofSocket::SetBroadcast(BOOL bBroadcast)
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return FALSE;

		//Set broadcast option
		if(setsockopt(GetHandle(),
					  SOL_SOCKET,
					  SO_BROADCAST,
					  (char*)&bBroadcast,
					  sizeof(bBroadcast))==GetErrorCode())
		{
			//Check for options error
			SetLastError("SetBroadcast");

			//Exit
			return FALSE;
		}	

		return TRUE;
	}
	ERROR_HANDLER_RETURN("SetBroadcast",FALSE)
}

unsigned char CSpoofSocket::GetTTL()const
{
	return m_ucTTL;
}

IP CSpoofSocket::GetSourceAddress()const
{
	return m_ulSourceAddress;
}

BOOL CSpoofSocket::HasOptions()const
{
	return m_bOptions;
}

unsigned char CSpoofSocket::GetProtocol()const
{
	return m_ucProtocol;
}

BOOL CSpoofSocket::SetReceiveTimeout(unsigned long ulMS)
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return FALSE;

		//Set it
		if(setsockopt(GetHandle(),
					  SOL_SOCKET,
					  SO_RCVTIMEO,
					  (char*)&ulMS,
					  sizeof(ulMS))==GetErrorCode())
		{
			//Check for options error
			SetLastError("SetReceiveTimeout");

			//Exit
			return FALSE;
		}
		else
			return TRUE;
	}
	ERROR_HANDLER_RETURN("SetReceiveTimeout",FALSE)
}

bool CSpoofSocket::operator==(const CSpoofSocket& rSocket)const
{
	return m_aSpoofSocket==rSocket.m_aSpoofSocket;
}

bool CSpoofSocket::operator<(const CSpoofSocket& rSocket)const
{
	return m_aSpoofSocket<rSocket.m_aSpoofSocket;
}

void CSpoofSocket::SetConnectedTo(const sockaddr_in& rAddress)
{
	m_aConnectedTo=rAddress;
}

KOMODIA_NAMESPACE_END

